from dataclasses import dataclass
from typing import Any

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.multi_analyte_profiling.benchling._2024._09.multi_analyte_profiling import (
    AnalyteAggregateDocument,
    AnalyteDocumentItem,
    CalculatedDataAggregateDocument,
    CalculatedDataDocumentItem,
    CalibrationAggregateDocument,
    CalibrationDocumentItem,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    ErrorAggregateDocument,
    ErrorDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    MultiAnalyteProfilingAggregateDocument,
    MultiAnalyteProfilingDocumentItem,
    SampleDocument,
    StatisticDimensionAggregateDocument,
    StatisticDimensionDocumentItem,
    StatisticsAggregateDocument,
    StatisticsDocumentItem,
    TQuantityValueModel,
)
from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMicroliter,
    TQuantityValueNumber,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TStatisticDatumRole,
)
from allotropy.allotrope.models.shared.definitions.units import Unitless
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
)
from allotropy.parsers.utils.values import quantity_or_none


@dataclass(frozen=True)
class StatisticDimension:
    value: float
    unit: str
    statistic_datum_role: TStatisticDatumRole


@dataclass(frozen=True)
class StatisticsDocument:
    statistical_feature: str
    statistic_dimensions: list[StatisticDimension]


@dataclass(frozen=True)
class Analyte:
    identifier: str
    name: str
    assay_bead_identifier: str
    assay_bead_count: float
    statistics: list[StatisticsDocument] | None = None
    statistic_datum_role: TStatisticDatumRole | None = None
    fluorescence: float | None = None


@dataclass(frozen=True)
class Error:
    error: str
    feature: str | None = None


@dataclass(frozen=True)
class Measurement:
    # Measurement metadata
    identifier: str
    sample_identifier: str
    location_identifier: str
    measurement_time: str

    # Settings
    assay_bead_count: float

    analytes: list[Analyte]
    errors: list[Error] | None = None

    # Optional metadata
    description: str | None = None
    sample_role_type: SampleRoleType | None = None
    well_plate_identifier: str | None = None

    # Optional settings
    sample_volume_setting: float | None = None
    dilution_factor_setting: float | None = None
    minimum_assay_bead_count_setting: float | None = None
    detector_gain_setting: str | None = None

    # custom
    measurement_custom_info: dict[str, Any] | None = None
    sample_custom_info: dict[str, Any] | None = None
    device_control_custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    measurements: list[Measurement]
    plate_well_count: float
    analyst: str | None
    analytical_method_identifier: str | None = None
    experimental_data_identifier: str | None = None
    method_version: str | None = None
    container_type: str | None = None
    experiment_type: str | None = None
    custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class Calibration:
    name: str
    time: str
    report: str | None = None


@dataclass(frozen=True)
class Metadata:
    file_name: str
    asm_file_identifier: str
    unc_path: str
    device_type: str
    data_system_instance_identifier: str
    model_number: str | None = None
    software_name: str | None = None
    software_version: str | None = None
    equipment_serial_number: str | None = None
    firmware_version: str | None = None
    product_manufacturer: str | None = None

    calibrations: list[Calibration] | None = None
    custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]
    calculated_data: list[CalculatedDocument] | None = None


class Mapper(SchemaMapper[Data, Model]):
    MANIFEST = "http://purl.allotrope.org/manifests/multi-analyte-profiling/BENCHLING/2024/09/multi-analyte-profiling.manifest"

    def map_model(self, data: Data) -> Model:
        return Model(
            multi_analyte_profiling_aggregate_document=MultiAnalyteProfilingAggregateDocument(
                device_system_document=add_custom_information_document(
                    DeviceSystemDocument(
                        model_number=data.metadata.model_number,
                        product_manufacturer=data.metadata.product_manufacturer,
                        firmware_version=data.metadata.firmware_version,
                        equipment_serial_number=data.metadata.equipment_serial_number,
                        calibration_aggregate_document=self._get_calibration_aggregate_document(
                            data.metadata.calibrations
                        ),
                    ),
                    custom_info_doc=data.metadata.custom_info,
                ),
                data_system_document=DataSystemDocument(
                    data_system_instance_identifier=data.metadata.data_system_instance_identifier,
                    file_name=data.metadata.file_name,
                    UNC_path=data.metadata.unc_path,
                    software_name=data.metadata.software_name,
                    software_version=data.metadata.software_version,
                    ASM_file_identifier=data.metadata.asm_file_identifier,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                multi_analyte_profiling_document=[
                    self._get_technique_document(measurement_group, data.metadata)
                    for measurement_group in data.measurement_groups
                ],
                calculated_data_aggregate_document=self._get_calculated_data_aggregate_document(
                    data.calculated_data
                ),
            ),
            field_asm_manifest=self.MANIFEST,
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup, metadata: Metadata
    ) -> MultiAnalyteProfilingDocumentItem:
        return MultiAnalyteProfilingDocumentItem(
            analyst=measurement_group.analyst,
            measurement_aggregate_document=add_custom_information_document(
                MeasurementAggregateDocument(
                    experiment_type=measurement_group.experiment_type,
                    analytical_method_identifier=measurement_group.analytical_method_identifier,
                    method_version=measurement_group.method_version,
                    experimental_data_identifier=measurement_group.experimental_data_identifier,
                    container_type=measurement_group.container_type,
                    plate_well_count=quantity_or_none(
                        TQuantityValueNumber, measurement_group.plate_well_count
                    ),
                    measurement_document=[
                        self._get_measurement_document(measurement, metadata)
                        for measurement in measurement_group.measurements
                    ],
                ),
                custom_info_doc=measurement_group.custom_info,
            ),
        )

    def _get_measurement_document(
        self, measurement: Measurement, metadata: Metadata
    ) -> MeasurementDocumentItem:
        measurement_document = MeasurementDocumentItem(
            measurement_identifier=measurement.identifier,
            measurement_time=self.get_date_time(measurement.measurement_time),
            sample_document=add_custom_information_document(
                self._get_sample_document(measurement),
                custom_info_doc=measurement.sample_custom_info,
            ),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    add_custom_information_document(
                        DeviceControlDocumentItem(
                            device_type=metadata.device_type,
                            sample_volume_setting=quantity_or_none(
                                TQuantityValueMicroliter,
                                measurement.sample_volume_setting,
                            ),
                            dilution_factor_setting=quantity_or_none(
                                TQuantityValueUnitless,
                                measurement.dilution_factor_setting,
                            ),
                            detector_gain_setting=measurement.detector_gain_setting,
                            minimum_assay_bead_count_threshold_setting=quantity_or_none(
                                TQuantityValueNumber,
                                measurement.minimum_assay_bead_count_setting,
                            ),
                        ),
                        custom_info_doc=measurement.device_control_custom_info,
                    )
                ]
            ),
            assay_bead_count=TQuantityValueNumber(value=measurement.assay_bead_count),
            analyte_aggregate_document=AnalyteAggregateDocument(
                analyte_document=[
                    AnalyteDocumentItem(
                        analyte_identifier=analyte.identifier,
                        analyte_name=analyte.name,
                        assay_bead_identifier=analyte.assay_bead_identifier,
                        assay_bead_count=TQuantityValueNumber(
                            value=analyte.assay_bead_count
                        ),
                        fluorescence=quantity_or_none(
                            TQuantityValueRelativeFluorescenceUnit,
                            analyte.fluorescence,
                            has_statistic_datum_role=analyte.statistic_datum_role,
                        ),
                        statistics_aggregate_document=(
                            StatisticsAggregateDocument(
                                statistics_document=[
                                    StatisticsDocumentItem(
                                        statistical_feature=statistic.statistical_feature,
                                        statistic_dimension_aggregate_document=StatisticDimensionAggregateDocument(
                                            statistic_dimension_document=[
                                                StatisticDimensionDocumentItem(
                                                    statistical_value=TQuantityValueModel(
                                                        value=dimension.value,
                                                        unit=dimension.unit,
                                                        has_statistic_datum_role=dimension.statistic_datum_role,
                                                    ),
                                                )
                                                for dimension in statistic.statistic_dimensions
                                            ]
                                        ),
                                    )
                                    for statistic in analyte.statistics
                                ]
                            )
                            if analyte.statistics
                            else None
                        ),
                    )
                    for analyte in measurement.analytes
                ]
            ),
            error_aggregate_document=self._get_error_aggregate_document(
                measurement.errors
            ),
        )

        return add_custom_information_document(
            measurement_document,
            custom_info_doc=measurement.measurement_custom_info,
        )

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:
        return SampleDocument(
            sample_identifier=measurement.sample_identifier,
            location_identifier=measurement.location_identifier,
            description=measurement.description,
            sample_role_type=(
                measurement.sample_role_type.value
                if measurement.sample_role_type
                else None
            ),
            well_plate_identifier=measurement.well_plate_identifier,
        )

    def _get_calibration_aggregate_document(
        self, calibrations: list[Calibration] | None
    ) -> CalibrationAggregateDocument | None:
        if not calibrations:
            return None
        return CalibrationAggregateDocument(
            calibration_document=[
                CalibrationDocumentItem(
                    calibration_name=calibration.name,
                    calibration_report=calibration.report,
                    calibration_time=self.get_date_time(calibration.time),
                )
                for calibration in calibrations
            ]
        )

    def _get_error_aggregate_document(
        self, errors: list[Error] | None
    ) -> ErrorAggregateDocument | None:
        if not errors:
            return None

        return ErrorAggregateDocument(
            error_document=[
                ErrorDocumentItem(error=error.error, error_feature=error.feature)
                for error in errors
            ]
        )

    def _get_calculated_data_aggregate_document(
        self, calculated_data_items: list[CalculatedDocument] | None
    ) -> CalculatedDataAggregateDocument | None:
        if not calculated_data_items:
            return None

        return CalculatedDataAggregateDocument(
            calculated_data_document=[
                CalculatedDataDocumentItem(
                    calculated_data_identifier=calculated_data_item.uuid,
                    calculated_data_name=calculated_data_item.name,
                    calculation_description=calculated_data_item.description,
                    calculated_result=TQuantityValueModel(
                        value=calculated_data_item.value,
                        unit=calculated_data_item.unit or Unitless.unit,
                    ),
                    data_source_aggregate_document=DataSourceAggregateDocument(
                        data_source_document=[
                            DataSourceDocumentItem(
                                data_source_identifier=item.reference.uuid,
                                data_source_feature=item.feature,
                            )
                            for item in calculated_data_item.data_sources
                        ]
                    ),
                )
                for calculated_data_item in calculated_data_items
            ]
        )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, cast

import pandas as pd

from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.json import JsonData
from allotropy.parsers.utils.strict_xml_element import StrictXmlElement
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float_or_none,
    try_int_or_none,
)
from allotropy.types import DictType


def _extract_from_xml_data(
    xml_data: StrictXmlElement | DictType | None, key: str = "value"
) -> str | None:
    """Extract a value from either StrictXmlElement or dictionary structure.

    Unified function that handles both XML element and dictionary structures,
    supporting extraction of any key (value, min, max, etc.).

    Handles structures like:
    Dictionary:
    - {"value": "25"} -> "25"
    - {"value": {"#text": "25"}} -> "25"
    - {"value": {"@IsUndefined": "False", "#text": "25"}} -> "25"
    - {"min": "4", "max": "45"} -> "4" or "45"

    StrictXmlElement:
    - <element value="25"/> -> "25"
    - <element><value>25</value></element> -> "25"
    - <element><value IsUndefined="False">25</value></element> -> "25"
    - <element min="4" max="45"/> -> "4" or "45"

    Args:
        xml_data: Either a StrictXmlElement or dictionary structure
        key: The name of the key/element/attribute to extract (default: "value")

    Returns:
        The extracted value as a string, or None if not found
    """
    if xml_data is None:
        return None

    # Handle StrictXmlElement
    if isinstance(xml_data, StrictXmlElement):
        value = xml_data.get_attr_or_none(key)
        if value is not None:
            return str(value)

        child_element = xml_data.find_or_none(key)
        if child_element is not None:
            text_value = child_element.get_text_or_none()
            if text_value is not None:
                return str(text_value)

            attr_value = child_element.get_attr_or_none("value")
            if attr_value is not None:
                return str(attr_value)

        # If key is "value" and no value found, try getting text content directly
        if key == "value":
            text_value = xml_data.get_text_or_none()
            if text_value is not None:
                return str(text_value)

        return None

    if isinstance(xml_data, dict):
        value = xml_data.get(key)
        if value is None:
            return None

        # Handle different value types
        if isinstance(value, str | int | float):
            return str(value)
        elif isinstance(value, dict):
            # If value is a dict with #text, extract the text
            value_dict = cast(dict[str, Any], value)
            text_content = value_dict.get("#text")
            if text_content is not None:
                return str(text_content)
            return str(value_dict)
        else:
            return str(value)

    return None


@dataclass(frozen=True)
class LigandImmobilization:
    flow_cell_index: int
    ligand: str | None
    immob_file_path: str | None
    immob_date_time: str | None
    level: float | None
    comment: str | None


@dataclass(frozen=True)
class ChipData:
    sensor_chip_identifier: str
    sensor_chip_type: str | None
    number_of_flow_cells: int | None
    number_of_spots: int | None
    lot_number: str | None
    immobilizations: list[LigandImmobilization]
    custom_info: dict[str, Any]

    @staticmethod
    def create(chip_data: DictType) -> ChipData:
        json_data = JsonData(dict(chip_data))

        immobilizations: list[LigandImmobilization] = []
        # Collect entries like Ligand{fc},1, Level{fc},1, ImmobFile{fc},1, ImmobDate{fc},1, Comment{fc},1
        num_flow_cells = json_data.get(int, "NoFcs", 0)
        for flow_cell in range(1, num_flow_cells + 1):
            if flow_cell < 1:
                continue
            ligand = json_data.get(str, f"Ligand{flow_cell},1")
            immob_file_path = json_data.get(str, f"ImmobFile{flow_cell},1")
            immob_date_time = json_data.get(str, f"ImmobDate{flow_cell},1")
            level = json_data.get(float, f"Level{flow_cell},1")
            comment = json_data.get(str, f"Comment{flow_cell},1")
            immobilizations.append(
                LigandImmobilization(
                    flow_cell_index=flow_cell,
                    ligand=ligand,
                    immob_file_path=immob_file_path,
                    immob_date_time=immob_date_time,
                    level=level,
                    comment=comment,
                )
            )

        sensor_chip_identifier = json_data[str, "Id", "Chip ID not found"]
        sensor_chip_type = json_data.get(str, "Name")
        number_of_flow_cells = json_data.get(int, "NoFcs")
        number_of_spots = json_data.get(int, "NoSpots")
        lot_number = json_data.get(str, "LotNo")

        custom_info = {
            "display name": json_data.get(str, "DisplayName"),
            "IFC": json_data.get(str, "IFC"),
            "IFC Description": json_data.get(str, "IFCDesc"),
            "First Dock Date": json_data.get(str, "FirstDockDate"),
            "Last Use Time": json_data.get(str, "LastUseTime"),
            "Last Modified Time": json_data.get(str, "LastModTime"),
            "Number of Flow Cells": json_data.get(str, "NoFcs"),
            "Number of Spots": json_data.get(str, "NoSpots"),
        }

        # Add any remaining unread fields to preserve all data
        custom_info.update(json_data.get_unread())
        custom_info = {k: v for k, v in custom_info.items() if v is not None}

        return ChipData(
            sensor_chip_identifier=sensor_chip_identifier,
            sensor_chip_type=sensor_chip_type,
            number_of_flow_cells=number_of_flow_cells,
            number_of_spots=number_of_spots,
            lot_number=lot_number,
            immobilizations=immobilizations,
            custom_info=custom_info,
        )


@dataclass(frozen=True)
class DetectionConfig:
    unread_detection_data: dict[str, Any]
    detection: str | None = None
    detection_dual: str | None = None
    detection_multi: str | None = None
    flow_cell_single: str | None = None
    flow_cell_dual: str | None = None
    flow_cell_multi: str | None = None

    @staticmethod
    def create(detection: DictType) -> DetectionConfig:
        json_data = JsonData(dict(detection))

        return DetectionConfig(
            detection=json_data.get(str, "Detection"),
            detection_dual=json_data.get(str, "DetectionDual"),
            detection_multi=json_data.get(str, "DetectionMulti"),
            flow_cell_single=json_data.get(str, "FlowCellSingle"),
            flow_cell_dual=json_data.get(str, "FlowCellDual"),
            flow_cell_multi=json_data.get(str, "FlowCellMulti"),
            unread_detection_data=json_data.get_unread(),
        )


@dataclass(frozen=True)
class RunMetadata:
    analyst: str | None = None
    compartment_temperature: float | None = None
    baseline_flow: float | None = None
    data_collection_rate: float | None = None
    molecule_weight_unit: str | None = None
    detection_config: DetectionConfig | None = None
    buffer_volume: float | None = None
    rack_temperature_min: float | None = None
    rack_temperature_max: float | None = None
    analysis_temperature: float | None = None
    prime: bool | None = None
    normalize: bool | None = None
    timestamp: str | None = None

    @staticmethod
    def create(application_template_details: DictType | None) -> RunMetadata:
        if application_template_details is None:
            return RunMetadata()
        props = application_template_details.get("properties", {})
        return RunMetadata(
            analyst=props.get("User"),
            compartment_temperature=try_float_or_none(
                _extract_value_from_xml_element_or_dict(
                    application_template_details.get("RackTemperature", {})
                )
            ),
            baseline_flow=try_float_or_none(
                _extract_value_from_xml_element_or_dict(
                    application_template_details.get("BaselineFlow", {})
                )
            ),
            data_collection_rate=try_float_or_none(
                _extract_value_from_xml_element_or_dict(
                    application_template_details.get("DataCollectionRate", {})
                )
            ),
            molecule_weight_unit=_extract_value_from_xml_element_or_dict(
                application_template_details.get("MoleculeWeightUnit", {})
            ),
            detection_config=DetectionConfig.create(
                application_template_details.get("detection", {})
            ),
            buffer_volume=try_float_or_none(
                (application_template_details.get("prepare_run", {}) or {}).get(
                    "BufferAVolume"
                )
            ),
            rack_temperature_min=try_float_or_none(
                _extract_min_from_xml_data(
                    application_template_details.get("RackTemperature", {})
                )
                or application_template_details.get("RackTemperatureMin")
            ),
            rack_temperature_max=try_float_or_none(
                _extract_max_from_xml_data(
                    application_template_details.get("RackTemperature", {})
                )
                or application_template_details.get("RackTemperatureMax")
            ),
            analysis_temperature=try_float_or_none(
                (application_template_details.get("system_preparations", {}) or {}).get(
                    "AnalTemp"
                )
            ),
            prime=(
                application_template_details.get("system_preparations", {}) or {}
            ).get("Prime"),
            normalize=(
                application_template_details.get("system_preparations", {}) or {}
            ).get("Normalize"),
            timestamp=props.get("Timestamp"),
        )


@dataclass(frozen=True)
class SystemInformation:
    application_name: str | None
    application_version: str | None
    user_name: str | None
    system_controller_identifier: str | None
    os_type: str | None
    os_version: str | None
    measurement_time: str | None
    unread_application_properties: dict[str, Any]
    measurement_aggregate_fields: dict[str, Any]

    @staticmethod
    def create(
        system_information: DictType | None, application_properties: DictType | None
    ) -> SystemInformation:
        system_info_data = JsonData(dict(system_information or {}))
        app_props_data = JsonData(dict(application_properties or {}))

        # Get measurement time from either source, preferring application properties
        measurement_time = app_props_data.get(str, "Timestamp") or system_info_data.get(
            str, "Timestamp"
        )

        # Extract specific fields for measurement aggregate custom info before skipping them
        measurement_aggregate_fields = {}
        target_fields = [
            "TemplateExtension",
            "EvaluationMethodIsOptional",
            "TypeName",
            "AllowPublish",
        ]

        for _field in target_fields:
            # Check both sources, preferring application properties
            value = app_props_data.get(str, _field) or system_info_data.get(str, _field)
            if value is not None:
                measurement_aggregate_fields[_field] = value

        # Read other fields that we want to skip to avoid JsonData warnings
        system_info_data.get(str, "SoftwareVersion")
        system_info_data.get(str, "Software")
        system_info_data.get(str, "User")
        system_info_data.get(
            str, "Timestamp"
        )  # Already read above but ensure both sources are marked
        app_props_data.get(str, "SoftwareVersion")
        app_props_data.get(str, "Software")
        app_props_data.get(str, "User")
        app_props_data.get(
            str, "Timestamp"
        )  # Already read above but ensure both sources are marked

        return SystemInformation(
            application_name=system_info_data.get(str, "Application"),
            application_version=system_info_data.get(str, "Version"),
            user_name=system_info_data.get(str, "UserName"),
            system_controller_identifier=system_info_data.get(
                str, "SystemControllerId"
            ),
            os_type=system_info_data.get(str, "OSType"),
            os_version=system_info_data.get(str, "OSVersion"),
            measurement_time=measurement_time,
            unread_application_properties=app_props_data.get_unread(
                skip={"HtmlPreview", "Timestamp"}
            ),
            measurement_aggregate_fields=measurement_aggregate_fields,
        )


@dataclass(frozen=True)
class CycleData:
    cycle_number: int
    sensorgram_data: pd.DataFrame
    report_point_data: pd.DataFrame | None = None

    @staticmethod
    def create(cycle_dict: DictType) -> CycleData:
        # Handle both old format (with DataFrames) and new format (with file paths)
        sensorgram_data = cycle_dict.get("sensorgram_data")
        if sensorgram_data is None:
            # If no DataFrame provided, create a minimal one for compatibility
            import pandas as pd

            sensorgram_data = pd.DataFrame(
                {
                    "Flow Cell Number": [1, 1],
                    "Cycle Number": [cycle_dict.get("cycle_number", 1)] * 2,
                    "Time (s)": [0.0, 1.0],
                    "Sensorgram (RU)": [0.0, 0.0],
                }
            )

        return CycleData(
            cycle_number=assert_not_none(
                cycle_dict.get("cycle_number"), "cycle_number"
            ),
            sensorgram_data=sensorgram_data,
            report_point_data=cycle_dict.get("report_point_data"),
        )


@dataclass(frozen=True)
class DipSweep:
    flow_cell: str
    sweep_row: str
    response: list[int]


@dataclass(frozen=True)
class DipData:
    count: int
    timestamp: str
    norm_data: list[DipSweep]
    raw_data: list[DipSweep]

    @staticmethod
    def create(dip_dict: DictType | None) -> DipData | None:
        if dip_dict is None:
            return None

        def _make(entries: list[DictType]) -> list[DipSweep]:
            return [
                DipSweep(
                    flow_cell=e["flow_cell"],
                    sweep_row=e["sweep_row"],
                    response=list(e["response"]),
                )
                for e in entries
            ]

        return DipData(
            count=int(dip_dict.get("count", 0)),
            timestamp=assert_not_none(dip_dict.get("timestamp"), "dip.timestamp"),
            norm_data=_make(dip_dict.get("norm_data", [])),
            raw_data=_make(dip_dict.get("raw_data", [])),
        )


@dataclass(frozen=True)
class FitQuality:
    chi2_value: float | None
    chi2_units: str | None

    @staticmethod
    def create(fq: DictType | None) -> FitQuality:
        if not fq or "Chi2" not in fq:
            return FitQuality(None, None)
        chi2 = fq["Chi2"]
        return FitQuality(try_float_or_none(chi2.get("value")), chi2.get("units"))


@dataclass(frozen=True)
class Parameter:
    name: str
    value: float | None
    error: float | None
    units: str | None


@dataclass(frozen=True)
class CalculatedValue:
    name: str
    value: float | None
    units: str | None


@dataclass(frozen=True)
class KineticResult:
    fit_quality: FitQuality
    parameters: list[Parameter] = field(default_factory=list)
    calculated: list[CalculatedValue] = field(default_factory=list)

    @staticmethod
    def create(result_dict: DictType) -> KineticResult:
        fit_quality = FitQuality.create(result_dict.get("fit_quality"))
        params = [
            Parameter(
                name=key,
                value=try_float_or_none(val.get("value")),
                error=try_float_or_none(val.get("error")),
                units=val.get("units"),
            )
            for key, val in (result_dict.get("parameters", {}) or {}).items()
        ]
        calcs = [
            CalculatedValue(
                name=key,
                value=try_float_or_none(val.get("value")),
                units=val.get("units"),
            )
            for key, val in (result_dict.get("calculated", {}) or {}).items()
        ]
        return KineticResult(
            fit_quality=fit_quality, parameters=params, calculated=calcs
        )


@dataclass(frozen=True)
class KineticAnalysis:
    results_by_identifier: dict[str, KineticResult]

    @staticmethod
    def create(ka_dict: DictType | None) -> KineticAnalysis | None:
        if not ka_dict:
            return None
        return KineticAnalysis(
            results_by_identifier={
                k: KineticResult.create(v) for k, v in ka_dict.items()
            }
        )


@dataclass(frozen=True)
class Data:
    run_metadata: RunMetadata
    chip_data: ChipData
    system_information: SystemInformation
    total_cycles: int
    cycle_data: list[CycleData]
    dip: DipData | None
    kinetic_analysis: KineticAnalysis | None
    sample_data: Any | None
    application_template_details: DictType | None = None

    @staticmethod
    def create(intermediate_structured_data: DictType) -> Data:
        app_details: DictType | None = intermediate_structured_data.get(
            "application_template_details"
        )
        system_info: DictType | None = intermediate_structured_data.get(
            "system_information"
        )
        chip: DictType = intermediate_structured_data.get(
            "chip",
            {
                "Id": "N/A",
                "Name": None,
                "NoFcs": 4,
                "NoSpots": None,
                "LotNo": None,
            },
        )
        cycles_raw: list[DictType] = intermediate_structured_data.get("cycle_data", [])
        return Data(
            run_metadata=RunMetadata.create(app_details),
            chip_data=ChipData.create(chip),
            system_information=SystemInformation.create(
                system_info, (app_details or {}).get("properties")
            ),
            total_cycles=try_int_or_none(
                intermediate_structured_data.get("total_cycles")
            )
            or 0,
            cycle_data=[CycleData.create(c) for c in cycles_raw],
            dip=DipData.create(intermediate_structured_data.get("dip")),
            kinetic_analysis=KineticAnalysis.create(
                intermediate_structured_data.get("kinetic_analysis")
            ),
            sample_data=intermediate_structured_data.get("sample_data", NOT_APPLICABLE),
            application_template_details=app_details,
        )


# Convenience functions for common use cases
def _extract_value_from_xml_element_or_dict(
    xml_data: StrictXmlElement | DictType | None, value_name: str = "value"
) -> str | None:
    """Extract value from either StrictXmlElement or dictionary structure.

    This function serves as a bridge between the old xmltodict-based approach
    and the new StrictXmlElement approach.

    Args:
        xml_data: Either a StrictXmlElement or dictionary structure
        value_name: The name of the value element/attribute to look for

    Returns:
        The extracted value as a string, or None if not found
    """
    return _extract_from_xml_data(xml_data, value_name)


def _extract_min_from_xml_data(
    xml_data: StrictXmlElement | DictType | None,
) -> str | None:
    """Extract min value from either StrictXmlElement or dictionary structure."""
    return _extract_from_xml_data(xml_data, "min")


def _extract_max_from_xml_data(
    xml_data: StrictXmlElement | DictType | None,
) -> str | None:
    """Extract max value from either StrictXmlElement or dictionary structure."""
    return _extract_from_xml_data(xml_data, "max")



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass

from allotropy.calcdocs.config import CalculatedDataConfig
from allotropy.calcdocs.extractor import Element
from allotropy.calcdocs.view import Keys, ViewData
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
    DataSource,
)


@dataclass(frozen=True)
class CalculatedDataConfigWithOptional(CalculatedDataConfig):
    optional: bool = False

    def iter_data_sources(
        self,
        parent_keys: Keys,
        elements: list[Element],
        cache: dict[str, CalculatedDocument | None],
    ) -> Iterator[DataSource]:
        keys = self.view_data.filter_keys(parent_keys)
        item = self.view_data.get_item(keys)
        sub_keys_iterator = item.iter_keys() if isinstance(item, ViewData) else [Keys()]

        for sub_keys in sub_keys_iterator:
            new_keys = keys.append(sub_keys)
            if calc_doc := self.get_calc_doc(new_keys, cache):
                yield DataSource(
                    feature=calc_doc.name,
                    reference=calc_doc,
                    value=calc_doc.value,
                )
            elif self.optional:
                for sub_config in self.source_configs:
                    yield from sub_config.iter_data_sources(new_keys, elements, cache)



import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import (
    df_to_series_data,
    read_csv,
    read_excel,
    SeriesData,
)


class RevvityMatrixReader:
    SUPPORTED_EXTENSIONS = "csv,xlsx"
    data: pd.DataFrame
    headers: SeriesData

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        self.headers = SeriesData()

        if named_file_contents.extension == "csv":
            # Read the file normally
            df = read_csv(
                named_file_contents.contents, encoding=named_file_contents.encoding
            )
            # Check if this is the format with headers
            if self._has_header_format(df):
                self._extract_headers_and_data(df)
            else:
                df = df.dropna(axis=1, how="all")
                self.data = df

        else:
            df = read_excel(named_file_contents.contents)
            # Check if this is the format with headers
            if self._has_header_format(df):
                self._extract_headers_and_data(df)
            else:
                df = df.dropna(axis=1, how="all")
                self.data = df
                # For standard Excel format, fix viability if needed
                if "Viability" in df.columns:
                    # Reading a percent value (50%) in read_excel results in a decimal: 0.5
                    # Detect and adjust value back to 0-100%
                    first_row = df_to_series_data(df, 0)
                    viability = first_row[str, "Viability"]
                    if "%" not in first_row[str, "Viability"] and float(viability) < 1:
                        df["Viability"] = df["Viability"] * 100
                    first_row.get_unread()

    def _has_header_format(self, df: pd.DataFrame) -> bool:
        if len(df) > 0:
            cols = ",".join(map(str, df.columns))
            return "Plate Name" in cols
        return False

    def _extract_headers_and_data(self, df: pd.DataFrame) -> None:
        # Find where the data section starts (row with Row, Column)
        data_start_idx = None

        for i in range(len(df)):
            row_values = df.iloc[i].astype(str).tolist()
            row_str = ",".join([str(x) for x in row_values if pd.notna(x)])
            if "Row" in row_str and "Column" in row_str:
                data_start_idx = i
                break

        if data_start_idx is None:
            # If we can't find the proper markers, assume it's not a header format
            df = df.dropna(axis=1, how="all")
            self.data = df
            return

        # Extract column names first - these will be our first row of headers
        header_dict = {
            df.columns[0]: df.columns[1].split(":")[-1].strip()
            if ":" in df.columns[1]
            else df.columns[1]
        }

        # Extract headers from the rows above the data section
        for i in range(data_start_idx):
            row = df.iloc[i].tolist()
            if len(row) >= 2 and pd.notna(row[0]) and pd.notna(row[1]):
                key = str(row[0]).strip()
                value = str(row[1]).strip()
                if key in ["Dilution", "Assay Name"]:
                    value = value.split(":")[-1].strip() if ":" in value else value
                header_dict[key] = value

        clean_header_dict = {
            k: v for k, v in header_dict.items() if pd.notna(k) and pd.notna(v)
        }
        self.headers = SeriesData(pd.Series(clean_header_dict))

        # Extract the data section
        data_df = df.iloc[data_start_idx:].reset_index(drop=True)

        # Fix the type error by ensuring we convert the first row to column names properly
        first_row = data_df.iloc[0]
        # Convert the Series to a list of strings for the column names
        column_names = [str(x) for x in first_row.values]
        data_df.columns = pd.Index(column_names)

        data_df = data_df.iloc[1:].reset_index(drop=True)

        data_df = data_df.dropna(axis=1, how="all")
        self.data = data_df



from __future__ import annotations

import re
from typing import Any

from attr import dataclass
import numpy as np
import pandas as pd

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDalton,
    TQuantityValueMicroliterPerMinute,
    TQuantityValueNanomolar,
    TQuantityValueResponseUnit,
    TQuantityValueSecondTime,
)
from allotropy.allotrope.schema_mappers.adm.binding_affinity_analyzer.benchling._2024._12.binding_affinity_analyzer import (
    DeviceControlDocument,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.cytiva_biacore_insight import constants
from allotropy.parsers.cytiva_biacore_insight.cytiva_biacore_insight_reader import (
    CytivaBiacoreInsightReader,
)
from allotropy.parsers.utils.pandas import (
    df_to_series_data,
    drop_df_rows_while,
    map_rows,
    parse_header_row,
    SeriesData,
    split_dataframe,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    quantity_or_none,
)
from allotropy.types import DictType


def _get_table_from_dataframe(df: pd.DataFrame, split_on: str) -> pd.DataFrame:
    data_table = drop_df_rows_while(df, lambda row: row[0] != split_on)
    return parse_header_row(
        assert_not_none(data_table).replace(np.nan, None)
    ).reset_index(drop=True)


def _first_not_null_or_none(series: pd.Series[Any]) -> str | None:
    """Get the first non-null value from a Series or None."""
    return str(series[idx]) if (idx := series.first_valid_index()) is not None else None


@dataclass(frozen=True)
class RunMetadata:
    sensor_chip_id: str
    sensor_chip_type: str
    lot_number: str
    model_number: str
    equipment_serial_number: str
    measurement_time: str
    number_of_cycles: float
    data_collection_rate: float
    running_buffer: str
    measurement_end_time: str
    method_name: str
    compartment_temperature: float | None
    run_table: pd.DataFrame

    @staticmethod
    def create_runs(runs_data: pd.DataFrame) -> list[RunMetadata]:
        runs_data = runs_data.reset_index(drop=True)
        run, rest = split_dataframe(
            runs_data, lambda row: str(row[0]).startswith("Run")
        )
        runs = [run]
        while rest is not None:
            run, rest = split_dataframe(rest, lambda row: str(row[0]).startswith("Run"))
            runs.append(run)
        if len(runs) > 1:
            msg = "Instrument file contains multiple runs. Only single runs are supported at the moment."
            raise AllotropeConversionError(msg)
        return [RunMetadata.create(run) for run in runs if not run.empty]

    @staticmethod
    def create(run_data: pd.DataFrame) -> RunMetadata:
        def get_run_information(idx: int) -> SeriesData:
            idx1, idx2 = idx + 1, idx + 4
            rd1 = run_data.loc[idx1:idx2][[3, 4]].rename(columns={3: "n", 4: "v"})
            rd2 = run_data.loc[idx1:idx2][[6, 7]].rename(columns={6: "n", 7: "v"})
            rd3 = run_data.loc[idx1:idx2][[10, 11]].rename(columns={10: "n", 11: "v"})
            return df_to_series_data(parse_header_row(pd.concat((rd1, rd2, rd3)).T))

        run_table_start = 0
        chip_data = SeriesData()
        run_information = SeriesData()
        run_table = pd.DataFrame()
        compartment_temperature = None
        for index, row in run_data.iterrows():
            idx = int(str(index))
            if row[0] == "Chip information":
                chip_data = df_to_series_data(
                    parse_header_row(run_data.loc[idx + 1 : idx + 4][[0, 1]].T)
                )
                run_information = get_run_information(idx)
            elif row[0] == "Flow cell":
                run_table_start = idx
            elif row[0] == "Method summary":
                run_table = (
                    parse_header_row(run_data.loc[run_table_start : idx - 1])
                    .dropna(how="all")
                    .reset_index(drop=True)
                )
                run_table = run_table.loc[:, ~run_table.columns.str.contains("^nan$")]
            elif row[0] == "Set temperatures 1":
                if match := re.match(constants.COMPARTMENT_TEMP_REGEX, row[1]):
                    compartment_temperature = float(match.group(1))

        return RunMetadata(
            sensor_chip_id=chip_data[str, "Chip id"],
            sensor_chip_type=chip_data[str, "Chip type"],
            lot_number=chip_data[str, "Lot number"],
            model_number=run_information[str, "Instrument type"],
            equipment_serial_number=run_information[str, "Instrument id"],
            measurement_time=run_information[str, "Start"],
            number_of_cycles=run_information[float, "Cycles"],
            data_collection_rate=run_information[float, "Data collection rate"],
            running_buffer=run_information[str, "Running buffer"],
            measurement_end_time=run_information[str, "End"],
            method_name=run_information[str, "Method"],
            compartment_temperature=compartment_temperature,
            run_table=run_table,
        )


@dataclass(frozen=True)
class BiacoreInsightMetadata:
    file_name: str
    unc_path: str
    software_name: str
    software_version: str
    analyst: str
    analytical_method_id: str
    runs: list[RunMetadata]
    data_processing_document: DictType | None = None

    @staticmethod
    def create(reader: CytivaBiacoreInsightReader) -> BiacoreInsightMetadata:

        properties = reader.data["Properties"]
        properties, runs_data = split_dataframe(
            properties, lambda row: str(row[0]).startswith("Run")
        )
        if runs_data is None or runs_data.empty:
            msg = "No run data found in the properties section of instrument file."
            raise AllotropeConversionError(msg)
        for index, row in properties.iterrows():
            idx = int(str(index))
            if row[0] == "Evaluation":
                evaluation_data = df_to_series_data(
                    parse_header_row(properties.loc[idx + 1 : idx + 4][[0, 1]].T)
                )
            elif row[0] == "Software":
                software_data = df_to_series_data(
                    parse_header_row(properties.loc[idx + 1 : idx + 2][[0, 1]].T)
                )

        data_processing_doc = BiacoreInsightMetadata.get_data_processing_doc(reader)

        return BiacoreInsightMetadata(
            file_name=evaluation_data[str, "Name"],
            unc_path=evaluation_data[str, "Path"],
            software_name=software_data[str, "Name"],
            software_version=software_data[str, "Version"],
            analyst=evaluation_data[str, "Modified by"],
            analytical_method_id=evaluation_data[str, "Name"],
            runs=RunMetadata.create_runs(runs_data),
            data_processing_document=data_processing_doc,
        )

    @staticmethod
    def get_data_processing_doc(reader: CytivaBiacoreInsightReader) -> DictType | None:
        table_names = [
            "QC - Capture baseline",
            "QC - Capture level",
            "QC - Binding to reference",
        ]
        qc_tables = [table for table in table_names if table in reader.data]
        qc_df = None
        while qc_df is None and qc_tables:
            qc_table = qc_tables.pop(0)
            _, qc_df = split_dataframe(
                reader.data[qc_table],
                lambda row: row[0] == "Blank subtraction",
                include_split_row=True,
            )

        if qc_df is None or qc_df.empty:
            return None

        qc_metadata = df_to_series_data(
            parse_header_row(qc_df.dropna(how="all", axis=0).T[0:2])
        )
        metadata_keys = [
            "Blank subtraction",
            "Molecular weight adjustment",
            "Capture/ligand adjustment",
            "Adjustment for controls",
            "Curve analysis",
        ]
        return {key: qc_metadata.get(str, key) for key in metadata_keys}


@dataclass(frozen=True)
class KineticsData:
    acceptance_state: str | None = None
    curve_markers: str | None = None
    kinetics_model: str | None = None
    binding_on_rate_measurement_datum: float | None = None
    binding_off_rate_measurement_datum: float | None = None
    equilibrium_dissociation_constant: float | None = None
    maximum_binding_capacity: float | None = None
    kinetics_chi_squared: float | None = None
    tc: float | None = None

    @staticmethod
    def create(kinetics_data: SeriesData) -> KineticsData:
        return KineticsData(
            acceptance_state=kinetics_data.get(str, "Acceptance state"),
            curve_markers=kinetics_data.get(str, "Curve markers"),
            kinetics_model=kinetics_data.get(str, "Kinetics model"),
            binding_on_rate_measurement_datum=kinetics_data.get(float, "ka (1/Ms)"),
            binding_off_rate_measurement_datum=kinetics_data.get(float, "kd (1/s)"),
            equilibrium_dissociation_constant=kinetics_data.get(float, "KD (M)"),
            maximum_binding_capacity=kinetics_data.get(float, "Rmax (RU)"),
            kinetics_chi_squared=kinetics_data.get(float, "Kinetics Chi² (RU²)"),
            tc=kinetics_data.get(float, "tc"),
        )


class EvaluationKinetics:
    _data: dict[str, KineticsData]

    def __init__(self, kinetics_table: pd.DataFrame) -> None:
        def _get_key(row: pd.Series[Any]) -> str:
            return f'{row["Channel"]} {row["Capture 1 Solution"]} {row["Analyte 1 Solution"]}'

        self._data = {
            _get_key(row): KineticsData.create(SeriesData(row))
            for _, row in kinetics_table.iterrows()
        }

    def get_data(
        self,
        channel: int,
        capture_solution: str | None,
        analyte_solution: str | None,
    ) -> KineticsData:
        empty = KineticsData()
        if capture_solution is None or analyte_solution is None:
            return empty
        return self._data.get(f"{channel} {capture_solution} {analyte_solution}", empty)


@dataclass(frozen=True)
class ReportPointData:
    identifier: str
    identifier_role: str
    absolute_resonance: float
    time_setting: float
    relative_resonance: float | None
    step_name: str | None
    step_purpose: str | None
    window: float | None
    baseline: str | None
    lrsd: float | None
    slope: float | None
    standard_deviation: float | None

    @staticmethod
    def create(row: SeriesData) -> ReportPointData:
        return ReportPointData(
            identifier=random_uuid_str(),
            identifier_role=row[str, "Name"],
            absolute_resonance=row[float, "Absolute response (RU)"],
            time_setting=row[float, "Time (s)"],
            relative_resonance=row.get(float, "Relative response (RU)"),
            step_name=row.get(str, "Name"),
            step_purpose=row.get(str, "Step purpose"),
            window=row.get(float, "Window (s)"),
            baseline=row.get(str, "Baseline"),
            lrsd=row.get(float, "LRSD"),
            slope=row.get(float, "Slope (RU/s)"),
            standard_deviation=row.get(float, "Standard deviation"),
        )


@dataclass(frozen=True)
class MeasurementData:
    identifier: str
    sample_identifier: str
    method_name: str
    ligand_identifier: str | None
    device_control_document: list[DeviceControlDocument]
    sample_custom_info: dict[str, Any | None]
    kinetics: KineticsData
    report_point_data: list[ReportPointData]

    @staticmethod
    def create(
        channel_data: pd.DataFrame,
        metadata: BiacoreInsightMetadata,
        evaluation_kinetics: EvaluationKinetics,
    ) -> MeasurementData:
        identifier = random_uuid_str()
        run_metadata = metadata.runs[0]
        first_row_data = SeriesData(channel_data.iloc[0])
        run = first_row_data[int, "Run"]
        cycle_number = first_row_data[int, "Cycle"]
        channel = first_row_data[int, "Channel"]

        run_table = run_metadata.run_table
        device_control_document = []
        for flow_cell, flow_cell_data in channel_data.groupby("Flow cell"):
            run_data = run_table[
                (run_table["Flow cell"].astype(str) == str(flow_cell))
                & (run_table["Channel"] == channel)
            ]
            run_info = SeriesData() if run_data.empty else df_to_series_data(run_data)
            # All the info needed in the device control document is the same for all flow cell data
            flow_cell_info = SeriesData(flow_cell_data.iloc[0])
            device_control_document.append(
                DeviceControlDocument(
                    device_type=constants.DEVICE_TYPE,
                    flow_cell_identifier=str(flow_cell),
                    sample_temperature_setting=flow_cell_info.get(
                        float, "Temperature (°C)"
                    ),
                    device_control_custom_info={
                        "Analyte 1 Contact time": quantity_or_none(
                            TQuantityValueSecondTime,
                            flow_cell_info.get(float, "Analyte 1 Contact time (s)"),
                        ),
                        "Analyte 1 Dissociation time": quantity_or_none(
                            TQuantityValueSecondTime,
                            flow_cell_info.get(
                                float, "Analyte 1 Dissociation time (s)"
                            ),
                        ),
                        "Analyte 1 Flow rate": quantity_or_none(
                            TQuantityValueMicroliterPerMinute,
                            flow_cell_info.get(float, "Analyte 1 Flow rate (µl/min)"),
                        ),
                        "Regeneration 1 Contact time": quantity_or_none(
                            TQuantityValueSecondTime,
                            flow_cell_info.get(
                                float, "Regeneration 1 Contact time (s)"
                            ),
                        ),
                        "Regeneration 1 Flow rate": quantity_or_none(
                            TQuantityValueMicroliterPerMinute,
                            flow_cell_info.get(
                                float, "Regeneration 1 Flow rate (µl/min)"
                            ),
                        ),
                        "Included": run_info.get(str, "Included"),
                        "Sensorgram type": flow_cell_info.get(
                            str, "Sensorgram type", run_info.get(str, "Sensorgram type")
                        ),
                        "Level": quantity_or_none(
                            TQuantityValueResponseUnit,
                            run_info.get(float, "Level (RU)"),
                        ),
                    },
                )
            )

        capture_solution = _first_not_null_or_none(channel_data["Capture 1 Solution"])
        analyte_solution = first_row_data.get(str, "Analyte 1 Solution")

        return MeasurementData(
            identifier=identifier,
            sample_identifier=f"Run{run}_Cycle{cycle_number}",
            method_name=run_metadata.method_name,
            ligand_identifier=run_info.get(str, "Ligand"),
            device_control_document=device_control_document,
            sample_custom_info={
                "Run": run,
                "Cycle": cycle_number,
                "Channel": channel,
                "Analyte 1 Solution": analyte_solution,
                "Analyte 1 Plate id": first_row_data.get(str, "Analyte 1 Plate id"),
                "Analyte 1 Position": first_row_data.get(str, "Analyte 1 Position"),
                "Analyte 1 Control type": first_row_data.get(
                    str, "Analyte 1 Control type"
                ),
                "Regeneration 1 Solution": first_row_data.get(
                    str, "Regeneration 1 Solution"
                ),
                "Regeneration 1 Plate id": first_row_data.get(
                    str, "Regeneration 1 Plate id"
                ),
                "Regeneration 1 Position": first_row_data.get(
                    str, "Regeneration 1 Position"
                ),
                "Regeneration 1 Control type": first_row_data.get(
                    str, "Regeneration 1 Control type"
                ),
                # Capture data is not reported for the Reference flow cell,
                # so we have to look for the first non-null value
                "Capture 1 Solution": capture_solution,
                "Capture 1 Plate id": _first_not_null_or_none(
                    channel_data["Capture 1 Plate id"]
                ),
                "Capture 1 Position": _first_not_null_or_none(
                    channel_data["Capture 1 Position"]
                ),
                "Capture 1 Control type": _first_not_null_or_none(
                    channel_data["Capture 1 Control type"]
                ),
                "Analyte 1 Concentration": quantity_or_none(
                    TQuantityValueNanomolar,
                    first_row_data.get(float, "Analyte 1 Concentration (nM)"),
                ),
                "Analyte 1 Molecular weight": quantity_or_none(
                    TQuantityValueDalton,
                    first_row_data.get(float, "Analyte 1 Molecular weight (Da)"),
                ),
            },
            kinetics=evaluation_kinetics.get_data(
                channel, capture_solution, analyte_solution
            ),
            report_point_data=map_rows(channel_data, ReportPointData.create),
        )


@dataclass(frozen=True)
class Data:
    metadata: BiacoreInsightMetadata
    cycles: dict[int, list[MeasurementData]]

    @staticmethod
    def create(reader: CytivaBiacoreInsightReader) -> Data:
        metadata = BiacoreInsightMetadata.create(reader)
        return Data(
            metadata=metadata,
            cycles={
                cycle: Data.create_measurements_for_cycle(cycle, metadata, reader)
                for cycle in range(1, int(metadata.runs[0].number_of_cycles) + 1)
            },
        )

    @staticmethod
    def create_measurements_for_cycle(
        cycle_number: int,
        metadata: BiacoreInsightMetadata,
        reader: CytivaBiacoreInsightReader,
    ) -> list[MeasurementData]:
        # Report point table is the entry point
        report_point_table = _get_table_from_dataframe(
            reader.data["Report point table"], split_on="Run"
        )
        cycle_data = report_point_table[report_point_table["Cycle"] == cycle_number]
        evaluation_kinetics_table = _get_table_from_dataframe(
            reader.data["Evaluation - Kinetics"], split_on="Group"
        )
        evaluation_kinetics = EvaluationKinetics(evaluation_kinetics_table)

        return [
            MeasurementData.create(channel_data, metadata, evaluation_kinetics)
            for _, channel_data in cycle_data.groupby("Channel")
        ]



import xml.etree.ElementTree as ET  # noqa: N817

from allotropy.allotrope.models.adm.flow_cytometry.benchling._2025._03.flow_cytometry import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.flow_cytometry.benchling._2025._03.flow_cytometry import (
    Data,
    Mapper,
)
from allotropy.exceptions import AllotropeParsingError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.bd_biosciences_facsdiva.bd_biosciences_facsdiva_structure import (
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.bd_biosciences_facsdiva.constants import DISPLAY_NAME
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.strict_xml_element import StrictXmlElement
from allotropy.parsers.vendor_parser import VendorParser


class BDFACSDivaParser(VendorParser[Data, Model]):
    DISPLAY_NAME = DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "xml"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        try:
            root_element_et = ET.parse(  # noqa: S314
                named_file_contents.contents
            ).getroot()
            root_element = StrictXmlElement(root_element_et)
        except ET.ParseError as e:
            msg = f"There was an error when trying to read the xml file: {e}"
            raise AllotropeParsingError(msg) from e

        return Data(
            metadata=create_metadata(
                root_element, named_file_contents.original_file_path
            ),
            measurement_groups=create_measurement_groups(root_element),
        )



from allotropy.calcdocs.config import (
    CalcDocsConfig,
    CalculatedDataConfig,
    MeasurementConfig,
)
from allotropy.calcdocs.docs.example.parser_calcdocs.extractor import ExampleExtractor
from allotropy.calcdocs.docs.example.parser_calcdocs.views import MeanView, SumView
from allotropy.calcdocs.docs.example.parser_code.measurement import Measurement
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
)
from allotropy.parsers.utils.uuids import random_uuid_str


class ExampleParser:
    def read_data(self) -> list[Measurement]:
        return [
            Measurement(random_uuid_str(), "a", "x", 1.0, 3.0, 5.0),
            Measurement(random_uuid_str(), "a", "x", 2.0, 3.0, 5.0),
            Measurement(random_uuid_str(), "a", "x", 3.0, 7.0, 5.0),
            Measurement(random_uuid_str(), "a", "x", 4.0, 7.0, 5.0),
            Measurement(random_uuid_str(), "a", "y", 5.0, 11.0, 13.0),
            Measurement(random_uuid_str(), "a", "y", 6.0, 11.0, 13.0),
            Measurement(random_uuid_str(), "a", "y", 7.0, 15.0, 13.0),
            Measurement(random_uuid_str(), "a", "y", 8.0, 15.0, 13.0),
            Measurement(random_uuid_str(), "b", "x", 9.0, 19.0, 21.0),
            Measurement(random_uuid_str(), "b", "x", 10.0, 19.0, 21.0),
            Measurement(random_uuid_str(), "b", "x", 11.0, 23.0, 21.0),
            Measurement(random_uuid_str(), "b", "x", 12.0, 23.0, 21.0),
            Measurement(random_uuid_str(), "b", "y", 13.0, 27.0, 29.0),
            Measurement(random_uuid_str(), "b", "y", 14.0, 27.0, 29.0),
            Measurement(random_uuid_str(), "b", "y", 15.0, 31.0, 29.0),
            Measurement(random_uuid_str(), "b", "y", 16.0, 31.0, 29.0),
        ]

    def create_calculated_data(
        self, measurements: list[Measurement]
    ) -> list[CalculatedDocument]:
        elements = ExampleExtractor.get_elements(measurements)

        mean_view_data = MeanView().apply(elements)
        sum_view_data = MeanView(sub_view=SumView()).apply(elements)

        measurement_conf = MeasurementConfig(
            name="measurement",
            value="m",
        )

        sum_conf = CalculatedDataConfig(
            name="sumation",
            value="sum",
            view_data=sum_view_data,
            source_configs=(measurement_conf,),
        )

        mean_conf = CalculatedDataConfig(
            name="sum mean",
            value="mean",
            view_data=mean_view_data,
            source_configs=(sum_conf,),
        )

        configs = CalcDocsConfig(
            [
                sum_conf,
                mean_conf,
            ]
        )

        return [
            calc_doc
            for parent_calc_doc in configs.construct()
            for calc_doc in parent_calc_doc.iter_struct()
        ]

    def create_data(self) -> None:
        measurements = self.read_data()
        self.create_calculated_data(measurements)



from __future__ import annotations

from allotropy.calcdocs.extractor import Element
from allotropy.calcdocs.view import View


class ReportPointDataView(View):
    def __init__(self, sub_view: View | None = None):
        super().__init__(name="report_point_data", sub_view=sub_view)

    def sort_elements(self, elements: list[Element]) -> dict[str, list[Element]]:
        return {element.uuid: [element] for element in elements}



from allotropy.allotrope.models.adm.pcr.rec._2024._09.qpcr import Model
from allotropy.allotrope.schema_mappers.adm.pcr.rec._2024._09.qpcr import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_parser import (
    AppBioQuantStudioParser,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_reader import (
    AppBioQuantStudioXLSXReader,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis import constants
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_data_creator import (
    create_calculated_data,
    create_data,
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_reader import (
    DesignQuantstudioReader,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class AppBioQuantStudioDesignandanalysisParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "AppBio QuantStudio Design & Analysis"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = DesignQuantstudioReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def parse_rt_pcr(
        self, reader: DesignQuantstudioReader, original_file_path: str
    ) -> Data:
        rt_pcr_reader = AppBioQuantStudioXLSXReader(reader.contents, reader.header)
        data = AppBioQuantStudioParser().parse_data(rt_pcr_reader, original_file_path)

        data.metadata.data_system_instance_identifier = "N/A"
        data.metadata.device_type = constants.DEVICE_TYPE
        data.metadata.container_type = constants.CONTAINER_TYPE
        data.metadata.product_manufacturer = constants.PRODUCT_MANUFACTURER
        data.metadata.software_name = constants.RT_PCR_SOFTWARE_NAME
        data.metadata.software_version = constants.RT_PCR_SOFTWARE_VERSION
        data.metadata.custom_info = reader.header.get_unread()

        return data

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = DesignQuantstudioReader.create(named_file_contents)

        # special case to be handle by RT-Pcr quantstudio adapter
        if reader.has_sheet("Sample Setup"):
            return self.parse_rt_pcr(reader, named_file_contents.original_file_path)

        data = create_data(reader)
        return Data(
            create_metadata(
                data.header,
                named_file_contents.original_file_path,
                data.experiment_type,
            ),
            create_measurement_groups(data),
            create_calculated_data(data),
        )



from __future__ import annotations

from collections.abc import Callable, Iterable
from enum import Enum
import os
import re
import traceback
from typing import Any, Literal, overload, TypeVar
import warnings

from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.utils.iterables import get_first_not_none
from allotropy.parsers.utils.values import (
    assert_not_none,
    str_to_bool,
    try_float_or_none,
)

T = TypeVar("T", bool, float, int, str)
# Need to use this instead of type[T] to get mypy to realize primitive can be called to return T
Type_ = Callable[..., T]
KeyOrKeys = Iterable[str]
TypeAndKey = tuple[Type_[T], KeyOrKeys]
TypeAndKeyAndMsg = tuple[Type_[T], KeyOrKeys, str]
ValidateRaw = Callable[[Any], bool] | None


class JsonData:
    class ValidateRawMode(Enum):
        # Return None for key if raw value is None
        NOT_NONE = "NOT_NONE"

    NOT_NONE = ValidateRawMode.NOT_NONE

    @staticmethod
    def _validate_raw(v: Any, mode: ValidateRawMode | None) -> Any:
        if mode is JsonData.ValidateRawMode.NOT_NONE:
            return None if v is None else v
        return v

    def __init__(self, data: dict[str, Any] | None = None) -> None:
        self.data = {} if data is None else data
        self.read_keys: set[str] = set()
        self.errored = False
        self.creation_stack = traceback.extract_stack()

    def __del__(self) -> None:
        if self.errored:
            return
        # NOTE: this will be turned on by default when all callers have been updated to pass the warning.
        if unread_keys := set(self.data.keys()) - self.read_keys:
            if os.getenv("WARN_UNUSED_KEYS"):
                # Find the creation point in the stack (skip the JsonData.__init__ frame)
                creation_point = None
                for frame in reversed(self.creation_stack):
                    if frame.name != "__init__" or "json.py" not in frame.filename:
                        creation_point = (
                            f"{frame.filename}:{frame.lineno} in {frame.name}"
                        )
                        break

                creation_info = (
                    f" (created at {creation_point})" if creation_point else ""
                )

                warnings.warn(
                    f"JsonData went out of scope without reading all keys{creation_info}, unread: {sorted(unread_keys)}.",
                    stacklevel=2,
                )

    def _get_custom_key(self, key: str) -> float | str | None:
        if (float_value := self.get(float, key)) is not None:
            return float_value
        return self.get(str, key)

    def _key_matches(self, match_key: str, key: str) -> bool:
        if key == match_key:
            return True
        # "++" can cause re.compile to fail. Since it is never a valid regex expression
        # it is safe to escape it to prevent the error.
        match_key = match_key.replace("++", r"\+\+")
        return bool(re.fullmatch(match_key, key))

    def _get_matching_keys(self, key_or_keys: str | set[str]) -> set[str]:
        return {
            matched
            for regex_key in (
                key_or_keys if isinstance(key_or_keys, set) else {key_or_keys}
            )
            for matched in [
                k for k in self.data.keys() if self._key_matches(str(regex_key), str(k))
            ]
        }

    def get_custom_keys(
        self, key_or_keys: str | set[str]
    ) -> dict[str, float | str | None]:
        return {
            key: value
            for key in self._get_matching_keys(key_or_keys)
            if (
                value := self._validate_raw(
                    self._get_custom_key(key), JsonData.NOT_NONE
                )
            )
            is not None
        }

    def mark_read(self, key_or_keys: str | set[str]) -> None:
        self.read_keys |= self._get_matching_keys(key_or_keys)

    def get_unread(
        self, regex: str | None = None, skip: set[str] | None = None
    ) -> dict[str, Any]:
        skip = self._get_matching_keys(skip) if skip else set()
        # Mark explicitly skipped keys as "read". This not only covers the check below, but removes
        # them from the destructor warning.
        self.read_keys |= skip
        matching_keys = (
            self._get_matching_keys(regex) if regex else set(self.data.keys())
        )

        # Get all unread keys
        all_unread_keys = {
            key for key in (matching_keys - self.read_keys) if key in self.data
        }

        # Filter to only include keys with non-dict, non-list values
        unread_keys = {}
        for key in all_unread_keys:
            value = self.data[key]
            # Ignore nested dictionaries and lists - these should be handled explicitly
            if not isinstance(value, dict | list):
                unread_keys[key] = value

        if all_unread_keys:
            self.read_keys |= all_unread_keys

        return unread_keys

    def has_key(self, key: str) -> bool:
        return key in self.data

    def __getitem__(self, type_and_key: TypeAndKey[T] | TypeAndKeyAndMsg[T]) -> T:
        """
        Get a value of the specified type with the specified key, raising an error if the
        key is not found, or if the value cannot be converted to the type.
        If a third argument is provided, it is an error message to provide if they key is not found.

        value: float = json_data[float, key]
        value: str = json_data[str, key, f"Failed to find {key} in my data"]

        Parameters:
        type (str, int, float, bool): The datatype to return.
        key (str | Iterable[str]): The key (or iterable of keys) to use to lookup.
        msg (str | None): The message to give as an error if lookup or conversion fails.

        Returns:
        type: A value of the type provided.

        Raises
        AllotropeConversionError: If the lookup or conversion to type fails.
        """
        if isinstance(type_and_key, str):
            key = type_and_key
            value = self.data.get(key)
            if value is None:
                msg = f"Required field '{key}' not found in data"
                raise AllotropeConversionError(msg)
            return value

        # Handle the case where type, key (and optionally message) are provided
        if len(type_and_key) == 2:
            type_, key = type_and_key
            msg = None
        elif len(type_and_key) == 3:
            type_, key, msg = type_and_key
        else:
            msg = f"Invalid arguments to __getitem__: {type_and_key}"
            raise AllotropeConversionError(msg)

        try:
            result = assert_not_none(self.get(type_, key), str(key), msg=msg)
            return result
        except Exception as e:
            self.errored = True
            # Wrap any exception with AllotropeConversionError
            msg = msg or f"Error accessing field '{key}': {e!s}"
            raise AllotropeConversionError(msg) from e

    # This overload tells typing that if default is "None" then get might return None
    @overload
    def get(
        self,
        type_: Type_[T],
        key: KeyOrKeys,
        default: Literal[None] = None,
        validate: ValidateRawMode | None = None,
    ) -> T | None:
        ...

    # This overload tells typing that if default matches T, get will return T
    @overload
    def get(
        self,
        type_: Type_[T],
        key: KeyOrKeys,
        default: T,
        validate: ValidateRawMode | None = None,
    ) -> T:
        ...

    def get(
        self,
        type_: Type_[T],
        key: KeyOrKeys,
        default: T | None = None,
        validate: ValidateRawMode | None = None,
    ) -> T | None:
        """
        Get a value of the specified type with the specified key, returning a default value if the
        key is not found, or if the value cannot be converted to the type.
        If validate is provided, it will be used to validate the raw value from the data, returning None
        if the value fails validation.

        Parameters:
        type (str, int, float, bool): The datatype to return.
        key (str | Iterable[str]): The key (or iterable of keys) to use to lookup.
        default (type | None): The value to return if lookup or conversion fails (default=None).
        validate (ValidateRawMode): The method to use for validating raw value. Defaults to (value is not None).

        Returns:
        type: A value of the type provided or default value.
        """
        if not isinstance(key, str):
            return get_first_not_none(
                lambda k: self.get(type_, k, validate=validate), key
            )
        self.read_keys.add(key)
        raw_value = self._validate_raw(self.data.get(key), validate)
        try:
            # bool needs special handling to convert
            if type_ is bool:
                raw_value = (
                    None
                    if raw_value is None
                    else ("true" if str_to_bool(str(raw_value)) else "")
                )
            if type_ is float and isinstance(raw_value, str) and "%" in raw_value:
                raw_value = raw_value.strip("%")
            convert = try_float_or_none if type_ is float else type_
            # mypy can't figure out that try_float_or_none will only be used when type_ is float.
            value = None if raw_value is None else convert(raw_value)  # type: ignore[operator]
        except ValueError:
            value = None

        return default if value is None else value

    def get_keys_as_dict(
        self,
        field_mappings: dict[str, tuple[Any, str, Any | None]],
        skip: set[str] | None = None,
        include_unread: bool | None = None,
    ) -> dict[str, Any]:
        """
        Extract multiple fields from JsonData into a dictionary, with type conversion.

        Example:
            field_mappings = {
                "output_field1": (str, "input_field1", None),  # type, field name, default value
                "output_field2": (float, "input_field2", 0.0),
                "renamed_field": (int, "original_name", None),
            }
            result = json_data.get_keys_as_dict(field_mappings)

        Parameters:
        field_mappings: Mapping of output field names to tuples of (type, input field name, default)
                        If default value is None, the field will be omitted if not found
        skip: Set of field names to skip when including unread data
        include_unread: Whether to include unread fields in the result (keyword-only)

        Returns:
        Dictionary with extracted fields, filtered to remove None values
        """
        result = {}

        # Process explicit field mappings
        for output_field, (field_type, input_field, default) in field_mappings.items():
            value = self.get(field_type, input_field, default)
            if value is not None or (default is not None and field_type is not float):
                result[output_field] = value

        # Include unread fields if requested
        if include_unread:
            result.update(self.get_unread(skip=skip))

        # Filter out None values and empty strings
        return {k: v for k, v in result.items() if v is not None and v != ""}

    def get_nested(
        self,
        type_: Type_[T],
        path: list[str],
        default: T | None = None,
        validate: ValidateRawMode | None = None,
    ) -> T | None:
        """
        Get a nested value following the provided path, with type conversion.

        Parameters:
        type_ (Type_[T]): The type to convert the value to
        path (list[str]): List of keys representing the path to the nested value
        default (T | None): Default value to return if path doesn't exist or conversion fails
        validate (ValidateRawMode | None): Validation mode for the raw value

        Returns:
        T | None: The typed value or the default
        """
        if not path:
            return default

        current = self.data
        for _i, key in enumerate(path[:-1]):
            if not isinstance(current, dict) or key not in current:
                return default
            current = current.get(key, {})
            # Mark the intermediate path as read
            self.mark_read(key)

        if not isinstance(current, dict) or path[-1] not in current:
            return default

        # For the final key, use the regular get method
        temp_json_data = JsonData(current)
        value = temp_json_data.get(type_, path[-1], default, validate)

        # Mark the key as read in our object
        self.mark_read(path[-1])

        return value



from collections.abc import Iterator
import re

import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.values import assert_not_none


def _extract_spot_legend_spots(lines: list[str]) -> tuple[list[str], list[int]]:
    # Find spot legend if it exists
    for i, line in enumerate(lines):
        if (
            line.startswith("Spot")
            and i + 1 < len(lines)
            and lines[i + 1].startswith("Legend")
        ):
            end = i + 2
            while end < len(lines):
                if not lines[end].startswith(""):
                    break
                end += 1
            spot_legend = "\n".join(lines[i : end - 1])
            lines = lines[:i] + lines[end - 1 :]
            spots = [int(spot) for spot in re.findall(r"(\d+)", spot_legend)]
            return lines, sorted(spots)
    return lines, []


class MethodicalMindReader:
    SUPPORTED_EXTENSIONS = "txt"
    plate_headers: list[SeriesData]
    plate_data: list[pd.DataFrame]
    spots: list[int]

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        file_lines = read_to_lines(named_file_contents)
        reader = CsvReader(file_lines)

        self.plate_headers = []
        self.plate_data = []
        self.spots = []
        while reader.current_line_exists():
            lines = list(reader.pop_until(r"^=+Data=+$"))
            lines, self.spots = _extract_spot_legend_spots(lines)
            lines = [
                line.replace("\\t", "\t")
                for line in lines
                if "Digital_Signature" not in line
            ]
            kv_pairs = [line.split(":", 1) for line in lines if ":" in line]
            key_values = {
                key.strip(): value.strip() for key, value in kv_pairs if value.strip()
            }
            self.plate_headers.append(SeriesData(pd.Series(key_values)))

            # Pop data title line
            reader.pop()

            # Create dataframe of the plate, drop any columns with all Nan values
            data = assert_not_none(
                reader.lines_as_df(
                    list(reader.pop_until("===")), sep="\t", header=0, index_col=0
                ),
                "Luminescence data table",
            ).dropna(axis="columns", how="all")
            # There may be multiple rows per well row for additional measurements, and the extra rows are
            # not labelled. ffill the row label so that each row has the corresponding row label.
            data.index = pd.Index(data.index.to_series().ffill())

            data.index = data.index.astype(str).str.strip()
            data.index = pd.Index(data.index.to_series().replace("", pd.NA).ffill())
            data.columns = data.columns.astype(str).str.strip()
            self.plate_data.append(data)

            reader.drop_until("Stack ID")

        if (
            self.spots
            and len(self.plate_data) > 1
            and len(self.plate_data) != len(self.spots)
        ):
            msg = f"Unrecognized file format for MSD parser: spot legend reports {len(self.spots)} spots, but found {len(self.plate_data)} data tables."
            raise AllotropeConversionError(msg)

    def __iter__(self) -> Iterator[tuple[SeriesData, pd.DataFrame]]:
        yield from zip(self.plate_headers, self.plate_data, strict=True)



from pathlib import Path

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueHertz,
    TQuantityValueNumber,
    TQuantityValueSecondTime,
    TQuantityValueSquareResponseUnit,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.units import (
    ResponseUnitPerSecond,
    Unitless,
)
from allotropy.allotrope.schema_mappers.adm.binding_affinity_analyzer.benchling._2024._12.binding_affinity_analyzer import (
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ReportPoint,
)
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.cytiva_biacore_insight import constants
from allotropy.parsers.cytiva_biacore_insight.cytiva_biacore_insight_structure import (
    BiacoreInsightMetadata,
    Data,
    MeasurementData,
    ReportPointData,
)
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
    DataSource,
    Referenceable,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import quantity_or_none


def create_metadata(metadata: BiacoreInsightMetadata, file_path: str) -> Metadata:
    path = Path(file_path)
    run_metadata = metadata.runs[0]
    return Metadata(
        device_identifier=constants.DEVICE_IDENTIFIER,
        asm_file_identifier=path.with_suffix(".json").name,
        data_system_instance_identifier=NOT_APPLICABLE,
        model_number=run_metadata.model_number,
        sensor_chip_identifier=run_metadata.sensor_chip_id,
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        software_name=metadata.software_name,
        software_version=metadata.software_version,
        file_name=metadata.file_name,
        unc_path=metadata.unc_path,
        detection_type=constants.DETECTION_TYPE,
        equipment_serial_number=run_metadata.equipment_serial_number,
        compartment_temperature=run_metadata.compartment_temperature,
        sensor_chip_type=run_metadata.sensor_chip_type,
        lot_number=run_metadata.lot_number,
    )


def create_measurement_groups(data: Data) -> list[MeasurementGroup]:
    # Here we assume that the first run contains the relevant metadata for the measurement group.
    run_metadata = data.metadata.runs[0]
    return [
        MeasurementGroup(
            measurement_time=run_metadata.measurement_time,
            analytical_method_identifier=data.metadata.analytical_method_id,
            analyst=data.metadata.analyst,
            measurement_aggregate_custom_info={
                "number_of_cycles": TQuantityValueNumber(
                    value=run_metadata.number_of_cycles
                ),
                "data_collection_rate": TQuantityValueHertz(
                    value=run_metadata.data_collection_rate
                ),
                "running_buffer": run_metadata.running_buffer,
                "measurement_end_time": run_metadata.measurement_end_time,
            },
            measurements=_get_measurements(cycle_measurements, data.metadata),
        )
        for cycle_measurements in data.cycles.values()
    ]


def _get_measurements(
    measurement_data: list[MeasurementData], metadata: BiacoreInsightMetadata
) -> list[Measurement]:
    data_processing_document = dict(metadata.data_processing_document or {})

    return [
        Measurement(
            identifier=measurement.identifier,
            sample_identifier=measurement.sample_identifier,
            type_=MeasurementType.SURFACE_PLASMON_RESONANCE,
            method_name=measurement.method_name,
            ligand_identifier=measurement.ligand_identifier,
            device_control_document=measurement.device_control_document,
            sample_custom_info=measurement.sample_custom_info,
            binding_on_rate_measurement_datum__kon_=(
                measurement.kinetics.binding_on_rate_measurement_datum
            ),
            binding_off_rate_measurement_datum__koff_=(
                measurement.kinetics.binding_off_rate_measurement_datum
            ),
            equilibrium_dissociation_constant__kd_=(
                measurement.kinetics.equilibrium_dissociation_constant
            ),
            maximum_binding_capacity__rmax_=(
                measurement.kinetics.maximum_binding_capacity
            ),
            processed_data_custom_info=(
                {
                    "Kinetics Chi squared": quantity_or_none(
                        TQuantityValueSquareResponseUnit,
                        measurement.kinetics.kinetics_chi_squared,
                    ),
                    "tc": quantity_or_none(
                        TQuantityValueUnitless, measurement.kinetics.tc
                    ),
                }
            ),
            report_point_data=[
                ReportPoint(
                    identifier=rp.identifier,
                    identifier_role=rp.identifier_role,
                    absolute_resonance=rp.absolute_resonance,
                    time_setting=rp.time_setting,
                    relative_resonance=rp.relative_resonance,
                    custom_info={
                        "Step name": rp.step_name,
                        "Step purpose": rp.step_purpose,
                        "Window": quantity_or_none(TQuantityValueSecondTime, rp.window),
                        "Baseline": rp.baseline,
                    },
                )
                for rp in measurement.report_point_data
            ],
            data_processing_document={
                **data_processing_document,
                "Acceptance State": measurement.kinetics.acceptance_state,
                "Curve Markers": measurement.kinetics.curve_markers,
                "Kinetics Model": measurement.kinetics.kinetics_model,
            },
        )
        for measurement in measurement_data
    ]


def create_calculated_data(data: Data) -> list[CalculatedDocument]:
    calculated_data = []
    for cycle_measurements in data.cycles.values():
        for measurement in cycle_measurements:
            for rp in measurement.report_point_data:
                calculated_data.extend(_get_report_point_calc_data(rp))

    return calculated_data


def _get_report_point_calc_data(rp: ReportPointData) -> list[CalculatedDocument]:
    calc_data = []
    data_source = DataSource(
        feature="Absolute Response", reference=Referenceable(rp.identifier)
    )
    if rp.lrsd is not None:
        calc_data.append(
            CalculatedDocument(
                uuid=random_uuid_str(),
                name="LRSD",
                value=rp.lrsd,
                data_sources=[data_source],
                unit=Unitless.unit,
            )
        )
    if rp.slope is not None:
        calc_data.append(
            CalculatedDocument(
                uuid=random_uuid_str(),
                name="Slope",
                value=rp.slope,
                data_sources=[data_source],
                unit=ResponseUnitPerSecond.unit,
            )
        )
    if rp.standard_deviation is not None:
        calc_data.append(
            CalculatedDocument(
                uuid=random_uuid_str(),
                name="Standard deviation",
                value=rp.standard_deviation,
                data_sources=[data_source],
                unit=Unitless.unit,
            )
        )
    return calc_data



import re
import xml.etree.ElementTree as ET  # noqa: N817

from allotropy.allotrope.models.adm.flow_cytometry.benchling._2025._03.flow_cytometry import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.flow_cytometry.benchling._2025._03.flow_cytometry import (
    Data,
    Mapper,
)
from allotropy.exceptions import AllotropeParsingError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.flowjo.constants import DISPLAY_NAME
from allotropy.parsers.flowjo.flowjo_structure import (
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.strict_xml_element import StrictXmlElement
from allotropy.parsers.vendor_parser import VendorParser


def extract_flowjo_namespaces(root_element: ET.Element) -> dict[str, str]:
    attributes = root_element.attrib.items()
    namespaces_text = next(
        (value for key, value in attributes if re.search(r"{.*}", key) is not None),
        None,
    )
    if namespaces_text is None:
        return {}

    namespace_list = namespaces_text.split()
    namespaces = {
        "transforms": next(
            (ns for ns in namespace_list if "transformations" in ns), ""
        ),
        "data-type": next((ns for ns in namespace_list if "datatypes" in ns), ""),
        "gating": next((ns for ns in namespace_list if "gating" in ns), ""),
    }

    # Extract all xmlns namespace declarations
    for key, value in attributes:
        if key.startswith("xmlns:"):
            prefix = key[6:]  # Remove "xmlns:" prefix
            if prefix not in namespaces:  # Don't override existing mappings
                namespaces[prefix] = value

    # ElementTree doesn't preserve xmlns declarations as attributes, but we can detect namespace usage
    # and add common mappings manually
    for key, _ in attributes:
        if key.startswith("{http://www.w3.org/2001/XMLSchema-instance}"):
            namespaces["xsi"] = "http://www.w3.org/2001/XMLSchema-instance"
            break

    return namespaces


class FlowjoParser(VendorParser[Data, Model]):
    DISPLAY_NAME = DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "wsp"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        try:
            root_element_et = ET.parse(  # noqa: S314
                named_file_contents.contents
            ).getroot()
            # Extract namespaces and create StrictXmlElement
            namespaces = extract_flowjo_namespaces(root_element_et)
            root_element = StrictXmlElement(root_element_et, namespaces)
        except ET.ParseError as e:
            msg = f"There was an error when trying to read the xml file: {e}"
            raise AllotropeParsingError(msg) from e

        return Data(
            metadata=create_metadata(
                root_element, named_file_contents.original_file_path
            ),
            measurement_groups=create_measurement_groups(root_element),
        )



from enum import Enum

DISPLAY_NAME = "Beckman Coulter Biomek"
DEVICE_TYPE = "liquid handler"
PROBE_HEAD_DEVICE_TYPE = "liquid handler probe head"
PRODUCT_MANUFACTURER = "Beckman Coulter"
SOFTWARE_NAME = "BioMek Software"


class TransferStep(str, Enum):
    ASPIRATE = "Aspirate"
    DISPENSE = "Dispense"


class FileFormat(str, Enum):
    UNIFIED_PIPETTING = "unified_pipetting"
    UNIFIED_TRANSFER = "unified_transfer"
    PIPETTING = "pipetting"


# Column mappings for different file formats
UNIFIED_PIPETTING_COLUMNS = [
    "Time Stamp",
    "Pod",
    "Transfer Step",
    "Deck Position",
    "Labware Name",
    "Labware Barcode",
    "Well Index",
    "Sample Name",
    "Probe",
    "Amount",
    "Liquid Handling Technique",
]

UNIFIED_TRANSFER_COLUMNS = [
    "Time Stamp",
    "Pod",
    "Source Position",
    "Source Labware Name",
    "Source Labware Barcode",
    "Source Well Index",
    "Sample Name",
    "Destination Position",
    "Destination Labware Name",
    "Destination Labware Barcode",
    "Destination Well Index",
    "Destination Sample Name",
    "Probe",
    "Amount",
]

PIPETTING_COLUMNS = [
    "Time Stamp",
    "Pod",
    "Transfer Step",
    "Position",
    "Labware Name",
    "Labware Barcode",
    "Well Index",
    "Amount",
    "Liquid Handling Technique",
]



from __future__ import annotations

from pathlib import Path
import re
from typing import Any

import pandas as pd

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueHertz,
    TQuantityValueMilliliter,
    TQuantityValueResponseUnit,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.schema_mappers.adm.binding_affinity_analyzer.benchling._2024._12.binding_affinity_analyzer import (
    DeviceControlDocument,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ReportPoint,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.exceptions import AllotropeParsingError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.cytiva_biacore_t200_evaluation import constants
from allotropy.parsers.cytiva_biacore_t200_evaluation.constants import (
    DEVICE_IDENTIFIER,
    MODEL_NUMBER,
)
from allotropy.parsers.cytiva_biacore_t200_evaluation.cytiva_biacore_t200_evaluation_decoder import (
    decode_data,
)
from allotropy.parsers.cytiva_biacore_t200_evaluation.cytiva_biacore_t200_evaluation_structure import (
    _extract_value_from_xml_element_or_dict,
    CalculatedValue,
    CycleData,
    Data,
    KineticResult,
    Parameter,
    SystemInformation,
)
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import quantity_or_none, try_float_or_none


def _get_sensorgram_datacube(
    sensorgram_df: pd.DataFrame, *, cycle: int, flow_cell: str
) -> DataCube:
    # Extract all sensorgram data points
    time_vals = sensorgram_df["Time (s)"].astype(float).to_list()
    resp_vals = sensorgram_df["Sensorgram (RU)"].astype(float).to_list()
    return DataCube(
        label=f"Cycle{cycle}_FlowCell{flow_cell}",
        structure_dimensions=[
            DataCubeComponent(FieldComponentDatatype.double, "elapsed time", "s")
        ],
        structure_measures=[
            DataCubeComponent(FieldComponentDatatype.double, "resonance", "RU")
        ],
        dimensions=[time_vals],
        measures=[resp_vals],
    )


def create_metadata(data: Data, named_file_contents: NamedFileContents) -> Metadata:
    filepath = Path(named_file_contents.original_file_path)
    sys = data.system_information
    chip = data.chip_data
    # Fallback: if run metadata lacks compartment temp, try application_template_details.RackTemperature.value
    # Using the new bridge function that can handle both StrictXmlElement and dict
    rack_temp_val = _extract_value_from_xml_element_or_dict(
        (data.application_template_details or {}).get("RackTemperature", {})
    )
    # Additional fallback to system_preparations.RackTemp if present
    rack_temp_sys_prep = (
        (data.application_template_details or {}).get("system_preparations", {}) or {}
    ).get("RackTemp")
    compartment_temp = (
        data.run_metadata.compartment_temperature
        or try_float_or_none(rack_temp_val)
        or try_float_or_none(rack_temp_sys_prep)
    )

    return Metadata(
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        device_identifier=DEVICE_IDENTIFIER,
        asm_file_identifier=filepath.with_suffix(".json").name,
        model_number=MODEL_NUMBER,
        data_system_instance_identifier=sys.system_controller_identifier
        or NOT_APPLICABLE,
        file_name=filepath.name,
        unc_path=named_file_contents.original_file_path,
        software_name=sys.application_name,
        software_version=sys.application_version,
        detection_type=constants.SURFACE_PLASMON_RESONANCE,
        compartment_temperature=compartment_temp,
        sensor_chip_type=chip.sensor_chip_type,
        lot_number=chip.lot_number,
        sensor_chip_identifier=chip.sensor_chip_identifier,
        sensor_chip_custom_info=chip.custom_info,
        data_system_custom_info={
            "account identifier": sys.user_name,
            "operating system type": sys.os_type,
            "operating system version": sys.os_version,
            **sys.unread_application_properties,
        },
    )


def _extract_kinetic_parameter(
    kinetic_result: KineticResult | None, section: str, parameter_names: list[str]
) -> float | None:
    """Extract kinetic parameter value from KineticResult object."""
    if not kinetic_result:
        return None

    # Get the appropriate list based on section
    items: list[Parameter] | list[CalculatedValue]
    if section == "parameters":
        items = kinetic_result.parameters
    elif section == "calculated":
        items = kinetic_result.calculated
    else:
        return None

    # Search through the items for matching parameter names
    for item in items:
        if item.name.lower() in [name.lower() for name in parameter_names]:
            return float(item.value) if item.value is not None else None

    return None


def _extract_kinetic_parameter_error(
    kinetic_result: KineticResult | None, parameter_names: list[str]
) -> float | None:
    """Extract kinetic parameter error from KineticResult object."""
    if not kinetic_result:
        return None

    # Search through parameters for matching parameter names
    for item in kinetic_result.parameters:
        if item.name.lower() in [name.lower() for name in parameter_names]:
            return float(item.error) if item.error is not None else None

    return None


def _extract_chi2_value(kinetic_result: KineticResult | None) -> float | None:
    """Extract Chi2 value from KineticResult fit quality."""
    if not kinetic_result or not kinetic_result.fit_quality:
        return None

    return (
        float(kinetic_result.fit_quality.chi2_value)
        if kinetic_result.fit_quality.chi2_value is not None
        else None
    )


def _create_report_point(
    series_data: SeriesData,
    flow_cell_id: str,
    cycle_number: int,
    display_flow_cell_id: str | None = None,
) -> ReportPoint | None:
    """Create a single ReportPoint object from SeriesData."""
    try:
        time_setting = series_data.get(float, ["column1", "Time"], default=0.0)
        relative_resonance = series_data.get(
            float, ["column3", "Relative"], default=0.0
        )
        identifier_role = series_data.get(str, ["column4", "Role"], default="baseline")
        absolute_resonance = series_data.get(
            float, ["column5", "Absolute"], default=0.0
        )

        unread_data = series_data.get_unread()

        fc_id_for_display = display_flow_cell_id or flow_cell_id
        report_point_id = f"CYTIVA_BIACORE_T200_EVALUATION_RP_C{cycle_number}_FC{fc_id_for_display}_{random_uuid_str()}"

        custom_info: dict[str, dict[str, object]] = {
            "window": {"value": 5.0, "unit": "s"}
        }
        for key, value in unread_data.items():
            custom_info[key] = {"value": value}

        return ReportPoint(
            identifier=report_point_id,
            identifier_role=identifier_role,
            absolute_resonance=absolute_resonance,
            time_setting=time_setting,
            relative_resonance=relative_resonance,
            custom_info=custom_info,
        )
    except Exception:
        series_data.get_unread()
        return None


def _create_report_points_from_cycle_data(
    rp_df: pd.DataFrame | None,
    flow_cell_id: str,
    cycle_number: int,
    display_flow_cell_id: str | None = None,
) -> list[ReportPoint] | None:
    """Create ReportPoint objects from cycle report point data, filtered by flow cell."""
    if rp_df is None or rp_df.empty:
        return None

    filtered_df = rp_df
    if "Flow Cell Number" in rp_df.columns or "flow_cell" in rp_df.columns:
        # Try to filter by flow cell
        fc_col = (
            "Flow Cell Number" if "Flow Cell Number" in rp_df.columns else "flow_cell"
        )
        # Convert flow_cell_id to match the format in the DataFrame
        try:
            fc_filter_value = int(flow_cell_id)
            filtered_df = rp_df[rp_df[fc_col] == fc_filter_value]
        except (ValueError, KeyError):
            # If filtering fails, use all data (fallback)
            filtered_df = rp_df

    report_points = [
        rp
        for rp in map_rows(
            filtered_df,
            lambda series_data: _create_report_point(
                series_data, flow_cell_id, cycle_number, display_flow_cell_id
            ),
        )
        if rp is not None
    ]

    return report_points if report_points else None


def _create_measurements_for_cycle(data: Data, cycle: CycleData) -> list[Measurement]:
    sensorgram_df = cycle.sensorgram_data
    cycle_num = cycle.cycle_number

    if "Flow Cell Number" not in sensorgram_df.columns:
        sensorgram_df["Flow Cell Number"] = 1
    if "Cycle Number" not in sensorgram_df.columns:
        sensorgram_df["Cycle Number"] = cycle_num

    rp_df: pd.DataFrame | None = cycle.report_point_data

    measurements: list[Measurement] = []

    def _normalize_flow_cell_id(value: Any) -> str:
        s = str(value)
        # Don't normalize reference-subtracted flow cell IDs (e.g., "2-1", "3-1", "4-1")
        if "-" in s:
            return s
        # Only normalize pure numeric flow cell IDs
        m = re.match(r"\d+", s)
        return m.group(0) if m else s

    # Process all flow cells (including reference-subtracted ones like "2-1", "3-1", "4-1")
    for flow_cell, df_fc in sensorgram_df.groupby("Flow Cell Number"):
        fc_id = _normalize_flow_cell_id(flow_cell)
        display_fc_id = (
            fc_id  # Use the flow cell ID (preserves reference-subtracted format)
        )

        # Extract report points from cycle data (use base fc_id for filtering data, display_fc_id for identifiers)
        report_points: list[ReportPoint] | None = _create_report_points_from_cycle_data(
            rp_df, fc_id, cycle_num, display_fc_id
        )

        device_control_custom_info: dict[str, Any] = {
            "buffer volume": quantity_or_none(
                TQuantityValueMilliliter, data.run_metadata.buffer_volume
            ),
            "detection": (
                data.run_metadata.detection_config.detection
                if data.run_metadata.detection_config
                else None
            ),
            "detectiondual": (
                data.run_metadata.detection_config.detection_dual
                if data.run_metadata.detection_config
                else None
            ),
            "detectionmulti": (
                data.run_metadata.detection_config.detection_multi
                if data.run_metadata.detection_config
                else None
            ),
            "flowcellsingle": (
                data.run_metadata.detection_config.flow_cell_single
                if data.run_metadata.detection_config
                else None
            ),
            "flowcelldual": (
                data.run_metadata.detection_config.flow_cell_dual
                if data.run_metadata.detection_config
                else None
            ),
            "flowcellmulti": (
                data.run_metadata.detection_config.flow_cell_multi
                if data.run_metadata.detection_config
                else None
            ),
            "maximum operating temperature": quantity_or_none(
                TQuantityValueDegreeCelsius, data.run_metadata.rack_temperature_max
            ),
            "minimum operating temperature": quantity_or_none(
                TQuantityValueDegreeCelsius, data.run_metadata.rack_temperature_min
            ),
            "analysis temperature": quantity_or_none(
                TQuantityValueDegreeCelsius, data.run_metadata.analysis_temperature
            ),
            "prime": str(bool(data.run_metadata.prime)).lower()
            if data.run_metadata.prime is not None
            else None,
            "normalize": str(bool(data.run_metadata.normalize)).lower()
            if data.run_metadata.normalize is not None
            else None,
        }

        # Add any unread detection data to device_control_custom_info
        if (
            data.run_metadata.detection_config
            and data.run_metadata.detection_config.unread_detection_data
        ):
            device_control_custom_info.update(
                data.run_metadata.detection_config.unread_detection_data
            )
        # Add experimental data identifier per measurement via chip immobilization mapping
        try:
            fc_index = int(str(fc_id))
        except Exception:
            fc_index = None
        if fc_index is not None:
            for imm in data.chip_data.immobilizations:
                if imm.flow_cell_index == fc_index and imm.ligand:
                    device_control_custom_info = {
                        **device_control_custom_info,
                        "ligand identifier": imm.ligand,
                    }
                if imm.flow_cell_index == fc_index and imm.level is not None:
                    device_control_custom_info = {
                        **device_control_custom_info,
                        "level": quantity_or_none(
                            TQuantityValueResponseUnit, imm.level
                        ),
                    }
                    break

        # Extract kinetic analysis data for this specific flow cell
        # Match EvaluationItem identifier to flow cell identifier
        combined_kinetic_data = None
        if data.kinetic_analysis and data.kinetic_analysis.results_by_identifier:
            # Try to find the specific EvaluationItem for this flow cell
            # Flow cell IDs are typically "1", "2", "3", "4"
            # EvaluationItem IDs are typically "EvaluationItem1", "EvaluationItem2", etc.
            matching_eval_item = None

            # First, try direct mapping: flow cell "1" -> "EvaluationItem1"
            eval_item_key = f"EvaluationItem{fc_id}"
            if eval_item_key in data.kinetic_analysis.results_by_identifier:
                matching_eval_item = eval_item_key
            else:
                # If direct mapping fails, look for any EvaluationItem that might correspond to this flow cell
                # This could be enhanced with more sophisticated matching logic if needed
                for eval_key in data.kinetic_analysis.results_by_identifier.keys():
                    if fc_id in eval_key or eval_key.endswith(fc_id):
                        matching_eval_item = eval_key
                        break

            # Use only the matching EvaluationItem data for this flow cell
            if matching_eval_item:
                result = data.kinetic_analysis.results_by_identifier[matching_eval_item]
                combined_kinetic_data = result

        kinetic_data = combined_kinetic_data

        measurements.append(
            Measurement(
                identifier=random_uuid_str(),
                type_=MeasurementType.SURFACE_PLASMON_RESONANCE,
                sample_identifier=NOT_APPLICABLE,
                device_control_document=[
                    DeviceControlDocument(
                        device_type=constants.DEVICE_TYPE,
                        flow_cell_identifier=display_fc_id,
                        flow_rate=try_float_or_none(data.run_metadata.baseline_flow),
                        detection_type=constants.SURFACE_PLASMON_RESONANCE,
                        device_control_custom_info=device_control_custom_info,
                    )
                ],
                well_plate_identifier=(
                    (
                        (data.application_template_details or {}).get("racks", {}) or {}
                    ).get("_Rack1")
                ),
                sample_custom_info={
                    "rack2": (
                        (data.application_template_details or {}).get("racks", {}) or {}
                    ).get("_Rack2")
                },
                sensorgram_data_cube=_get_sensorgram_datacube(
                    df_fc, cycle=cycle_num, flow_cell=fc_id
                ),
                report_point_data=report_points,
                # Kinetic analysis fields
                binding_on_rate_measurement_datum__kon_=_extract_kinetic_parameter(
                    kinetic_data, "parameters", ["ka", "kon"]
                ),
                binding_off_rate_measurement_datum__koff_=_extract_kinetic_parameter(
                    kinetic_data, "parameters", ["kd", "koff"]
                ),
                equilibrium_dissociation_constant__kd_=_extract_kinetic_parameter(
                    kinetic_data, "calculated", ["Kd_M", "KD", "kd"]
                ),
                maximum_binding_capacity__rmax_=_extract_kinetic_parameter(
                    kinetic_data, "parameters", ["Rmax", "rmax"]
                ),
                # Attach custom kinetic analysis values for processed data custom info
                processed_data_custom_info={
                    "kinetics chi squared": {
                        "value": _extract_chi2_value(kinetic_data),
                        "unit": "(unitless)",
                    },
                    "ka error": {
                        "value": _extract_kinetic_parameter_error(
                            kinetic_data, ["ka", "kon"]
                        ),
                        "unit": "M-1s-1",
                    },
                    "kd error": {
                        "value": _extract_kinetic_parameter_error(
                            kinetic_data, ["kd", "koff"]
                        ),
                        "unit": "s^-1",
                    },
                    "Rmax error": {
                        "value": _extract_kinetic_parameter_error(
                            kinetic_data, ["Rmax", "rmax"]
                        ),
                        "unit": "RU",
                    },
                },
            )
        )

    return measurements


def create_measurement_groups(data: Data) -> list[MeasurementGroup]:
    sys = data.system_information
    # Prefer application template timestamp if present in run metadata
    if data.run_metadata.timestamp and not sys.measurement_time:
        sys = SystemInformation(
            application_name=sys.application_name,
            application_version=sys.application_version,
            user_name=sys.user_name,
            system_controller_identifier=sys.system_controller_identifier,
            os_type=sys.os_type,
            os_version=sys.os_version,
            measurement_time=data.run_metadata.timestamp,
            unread_application_properties=sys.unread_application_properties,
            measurement_aggregate_fields=sys.measurement_aggregate_fields,
        )
    # As a final fallback, look directly in application_template_details.properties
    if not sys.measurement_time and data.application_template_details:
        props = data.application_template_details.get("properties", {})
        ts = props.get("Timestamp")
        if ts:
            sys = SystemInformation(
                application_name=sys.application_name,
                application_version=sys.application_version,
                user_name=sys.user_name,
                system_controller_identifier=sys.system_controller_identifier,
                os_type=sys.os_type,
                os_version=sys.os_version,
                measurement_time=ts,
                unread_application_properties=sys.unread_application_properties,
                measurement_aggregate_fields=sys.measurement_aggregate_fields,
            )
    if not sys.measurement_time:
        msg = "Missing measurement time. Expected application_template_details.properties.Timestamp."
        raise AllotropeParsingError(msg)
    groups: list[MeasurementGroup] = []
    # Process all cycles to create one measurement document per cycle
    for cycle in data.cycle_data:
        measurements = _create_measurements_for_cycle(data, cycle)
        custom_info: dict[str, Any] = {
            "data collection rate": quantity_or_none(
                TQuantityValueHertz, data.run_metadata.data_collection_rate
            ),
            **sys.measurement_aggregate_fields,
        }
        # Add aggregate-level experimental data identifier for convenience (first measurement's FC)
        if measurements:
            # Derive from first measurement's flow cell
            first_fc = measurements[0].device_control_document[0].flow_cell_identifier
            first_fc_str = str(first_fc)
            # Check if the flow cell identifier is a valid integer (skip reference-subtracted IDs like "2-1")
            if first_fc_str.isdigit():
                fc_index = int(first_fc_str)
                for imm in data.chip_data.immobilizations:
                    if imm.flow_cell_index == fc_index and imm.immob_file_path:
                        custom_info = {
                            **custom_info,
                            "experimental data identifier": imm.immob_file_path,
                        }
                        break

        groups.append(
            MeasurementGroup(
                measurement_time=sys.measurement_time,
                measurements=measurements,
                experiment_type=None,
                analytical_method_identifier=None,
                analyst=(
                    data.run_metadata.analyst or data.system_information.user_name
                ),
                measurement_aggregate_custom_info=custom_info,
            )
        )
    return groups


def create_data(
    named_file_contents: NamedFileContents,
) -> tuple[Metadata, list[MeasurementGroup]]:
    intermediate = decode_data(named_file_contents)
    data = Data.create(intermediate)
    metadata = create_metadata(data, named_file_contents)
    groups = create_measurement_groups(data)
    return metadata, groups



from collections import defaultdict
from datetime import timedelta

from allotropy.allotrope.models.shared.definitions.custom import (
    TNullableQuantityValueGramPerLiter,
    TNullableQuantityValueMillimolePerLiter,
    TNullableQuantityValueUnitPerLiter,
)

# Measurements of a sample have different timestamps, typically spaced closely together (max diff observed - 9 min)
# Some result files have multiple sets of measurements, over multiple days.
# Measurements with a time difference greater than MAX_MEASUREMENT_TIME_GROUP_DIFFERENCE from the last
# measurement recorded will be put into separate groups.
MAX_MEASUREMENT_TIME_GROUP_DIFFERENCE = timedelta(hours=1)


MOLAR_CONCENTRATION_CLASSES: list[
    (
        type[TNullableQuantityValueMillimolePerLiter]
        | type[TNullableQuantityValueGramPerLiter]
        | type[TNullableQuantityValueUnitPerLiter]
    )
] = [
    TNullableQuantityValueMillimolePerLiter,
    TNullableQuantityValueGramPerLiter,
    TNullableQuantityValueUnitPerLiter,
]

MOLAR_CONCENTRATION_CLS_BY_UNIT = {cls.unit: cls for cls in MOLAR_CONCENTRATION_CLASSES}


INFO_HEADER = [
    "row type",
    "data processing time",
    "col3",
    "col4",
    "col5",
    "col6",
    "model number",
    "device serial number",
    "software version",
    "analyst",
]

DATA_HEADER_V6 = [
    "row type",
    "measurement time",
    "sample identifier",
    "batch identifier",
    "sample role type",
    "col6",
    "analyte name",
    "col8",
    "concentration unit",
    "flag",
    "concentration value",
    "col12",
    "col13",
]

DATA_HEADER_V5 = [
    "row type",
    "measurement time",
    "analyte name",
    "col4",
    "col5",
    "sample identifier",
    "sample role type",
    "concentration unit",
    "flag",
    "concentration value",
    "col10",
    "col11",
]


ANALYTES_LOOKUP = {
    "AC2B": "acetate",
    "AC2D": "acetate",
    "AQB": "alanine-glutamine",
    "AQD": "alanine-glutamine",
    "AQQB": "alanine-glutamine corrected",
    "NH3LB": "ammonia",
    "NH3B": "ammonia",
    "NH3D": "ammonia",
    "ARAB": "arabinose",
    "ARAD": "arabinose",
    "ASNLB": "asparagine",
    "ASNHB": "asparagine",
    "ASNHD": "asparagine",
    "ASPB": "aspartate",
    "ASPD": "aspartate",
    "CA2B": "calcium",
    "CA2D": "calcium",
    "CHO2B": "cholesterol",
    "CHO2D": "cholesterol",
    "ETOHB": "ethanol",
    "ETOHD": "ethanol",
    "FORB": "formate",
    "FORD": "formate",
    "GAL2B": "galactose",
    "GAL2D": "galactose",
    "GLC2L": "glucose",
    "GLC2B": "glucose",
    "GLC2D": "glucose",
    "GLC3D": "glucose",
    "GLC3B": "glucose",
    "GLC3L": "glucose",
    "GLU2B": "glutamate",
    "GLU2D": "glutamate",
    "GLN2B": "glutamine",
    "GLN2D": "glutamine",
    "GLYB": "glycerol",
    "GLYD": "glycerol",
    "GLYE": "glycerol",
    "FABLA": "Ig Fab human",
    "FABLB": "Ig Fab human",
    "FABHB": "Ig Fab human",
    "FABHD": "Ig Fab human",
    "IGGLB": "IgG human",
    "IGGHB": "IgG human",
    "IGGHD": "IgG human",
    "MIGLB": "IgG mouse",
    "MIGHB": "IgG mouse",
    "MIGHD": "IgG mouse",
    "FE2B": "iron",
    "FE2D": "iron",
    "LAC2B": "lactate",
    "LAC2D": "lactate",
    "LDH2B": "ldh",
    "LDH2D": "ldh",
    "MG2L": "magnesium",
    "MG2B": "magnesium",
    "MG2D": "magnesium",
    "NO3L": "nitrate",
    "NO3B": "nitrate",
    "NO3D": "nitrate",
    "PHO2L": "phosphate",
    "PHO2B": "phosphate",
    "PHO2D": "phosphate",
    "KB": "potassium",
    "KD": "potassium",
    "PYRB": "pyruvate",
    "PYRD": "pyruvate",
    "NAB": "sodium",
    "NAD": "sodium",
    "SUCB": "sucrose",
    "SUCD": "sucrose",
    "SUGLB": "sucrose corrected",
    "SUGLD": "sucrose corrected",
    "TP2B": "total protein",
    "TP2D": "total protein",
    "TP2LB": "total protein",
    "ODB": "optical_density",
    "ODD": "optical_density",
    "OSM2B": "osmolality",
}

DETECTION_KIT_LOOKUP = {
    "AC2B": "Standard",
    "AC2D": "High with Pre - Dilution",
    "AQB": "Standard",
    "AQD": "High with Pre - Dilution",
    "AQQB": "Standard",
    "NH3LB": "Low",
    "NH3B": "Standard",
    "NH3D": "High with Pre - Dilution",
    "ARAB": "Standard",
    "ARAD": "High with Pre - Dilution",
    "ASNLB": "Low",
    "ASNHB": "Standard",
    "ASNHD": "High with Pre - Dilution",
    "ASPB": "Standard",
    "ASPD": "High with Pre - Dilution",
    "CA2B": "Standard",
    "CA2D": "High with Pre - Dilution",
    "CHO2B": "Standard",
    "CHO2D": "High with Pre - Dilution",
    "ETOHB": "Standard",
    "ETOHD": "High with Pre - Dilution",
    "FORB": "Standard",
    "FORD": "High with Pre - Dilution",
    "GAL2B": "Standard",
    "GAL2D": "High with Pre - Dilution",
    "GLC2L": "Low",
    "GLC2B": "Standard",
    "GLC2D": "High with Pre - Dilution",
    "GLC3D": "High with Pre - Dilution",
    "GLC3B": "Standard",
    "GLC3L": "Low",
    "GLU2B": "Standard",
    "GLU2D": "High with Pre - Dilution",
    "GLN2B": "Standard",
    "GLN2D": "High with Pre - Dilution",
    "GLYB": "Low",
    "GLYD": "Standard",
    "GLYE": "High with Pre - Dilution",
    "FABLA": "Low",
    "FABLB": "Standard",
    "FABHB": "High",
    "FABHD": "High with Pre - Dilution",
    "IGGLB": "Low",
    "IGGHB": "Standard",
    "IGGHD": "High with Pre - Dilution",
    "MIGLB": "Low",
    "MIGHB": "Standard",
    "MIGHD": "High with Pre - Dilution",
    "FE2B": "Standard",
    "FE2D": "High with Pre - Dilution",
    "LAC2B": "Standard",
    "LAC2D": "High with Pre - Dilution",
    "LDH2B": "Standard",
    "LDH2D": "High with Pre - Dilution",
    "MG2L": "Low",
    "MG2B": "Standard",
    "MG2D": "High with Pre - Dilution",
    "NO3L": "Low",
    "NO3B": "Standard",
    "NO3D": "High with Pre - Dilution",
    "PHO2L": "Low",
    "PHO2B": "Standard",
    "PHO2D": "High with Pre - Dilution",
    "KB": "Standard",
    "KD": "High with Pre - Dilution",
    "PYRB": "Standard",
    "PYRD": "High with Pre - Dilution",
    "NAB": "Standard",
    "NAD": "High with Pre - Dilution",
    "SUCB": "Standard",
    "SUCD": "High with Pre - Dilution",
    "SUGLB": "Standard",
    "SUGLD": "High with Pre - Dilution",
    "TP2B": "Standard",
    "TP2D": "High with Pre - Dilution",
    "TP2LB": "Low",
}

DETECTION_KIT_RANGE_LOOKUP = {
    "AC2B": "0.25 - 38.9 mmol/L",
    "AC2D": "14 - 389 mmol/L",
    "AQB": "0.1 - 10 mmol/L",
    "AQD": "1 - 100 mmol/L",
    "NH3LB": "0.0278 - 1.389 mmol/L",
    "NH3B": " 0.278 - 13.89 mmol/L",
    "NH3D": "5.56 - 277.8 mmol/L",
    "ARAB": "0.055 - 27.75 mmol/L",
    "ARAD": "0.55 - 277.5 mmol/L",
    "ASNLB": "0.1 - 1 mmol/L",
    "ASNHB": "0.5 - 9 mmol/L",
    "ASNHD": "2.5 - 45 mmol/L",
    "ASPB": "0.1 - 9 mmol/L",
    "ASPD": "1 - 45 mmol/L",
    "CA2B": "0.2 - 5 mmol/L",
    "CA2D": "2 - 50 mmol/L",
    "CHO2B": "0.025 - 2.07 mmol/L",
    "CHO2D": "0.25 - 20.7 mmol/L",
    "ETOHB": "11 - 220 mmol/L",
    "ETOHD": "220 - 4400 mmol/L",
    "FORB": "0.2 - 20 mmol/L",
    "FORD": "2 - 200 mmol/L",
    "GAL2B": "0.055 - 27.75 mmol/L",
    "GAL2D": "0.55 - 277.5 mmol/L",
    "GLC2L": "0.011 - 4.163 mmol/L",
    "GLC2B": "0.111 - 41.63 mmol/L",
    "GLC2D": "1.11 - 416.3 mmol/L",
    "GLC3D": "1.11 - 416.3 mmol/L",
    "GLC3B": "0.111 - 41.63 mmol/L",
    "GLC3L": "0.011 - 4.163 mmol/L",
    "GLU2B": "0.099 - 10.2 mmol/L",
    "GLU2D": "0.99 - 102 mmol/L",
    "GLN2B": "0.1 - 10.26 mmol/L",
    "GLN2D": "0.5 - 51.3 mmol/L",
    "GLYB": "0.1 - 10 mmol/L",
    "GLYD": "1 - 100 mmol/L",
    "GLYE": "10 - 1000 mmol/L ",
    "FABLA": "0.13 - 0.35 g/L",
    "FABLB": "0.01 - 0.4 g/L",
    "FABHB": "0.3 - 8 g/L",
    "FABHD": "6 - 40 g/L",
    "IGGLB": "0.000067 - 0.000533 mmol/L",
    "IGGHB": "0.000533 - 0.01067 mmol/L",
    "IGGHD": "0.002665 - 0.05335 mmol/L ",
    "MIGLB": "0.01 - 0.2 g/L",
    "MIGHB": "0.2 - 1.5 g/L",
    "MIGHD": "1 - 7.5 g/L",
    "FE2B": "0.0009 - 0.179 mmol/L",
    "FE2D": "0.009 - 1.79 mmol/L ",
    "LAC2B": "0.0444 - 15.55 mmol/L",
    "LAC2D": "0.444 - 155.5 mmol/L ",
    "LDH2B": "20 - 1000 U/L",
    "LDH2D": "200 - 10000 U/L",
    "MG2L": "0.01 - 0.25 mmol/L",
    "MG2B": "0.1 - 2.5 mmol/L",
    "MG2D": "2.3 - 25 mmol/L",
    "NO3L": "0.2 - 4 mmol/L",
    "NO3B": "1 - 25 mmol/L",
    "NO3D": "15 - 250 mmol/L",
    "PHO2L": "0.01 - 0.837 mmol/L",
    "PHO2B": "0.1 - 8.37 mmol/L",
    "PHO2D": "1.1 - 92.07 mmol/L",
    "KB": "2 - 28 mmol/L",
    "KD": "24 - 280 mmol/L",
    "PYRB": "0.2 - 8 mmol/L",
    "PYRD": "2 - 80 mmol/L",
    "NAB": "50 - 275 mmol/L",
    "NAD": "250 - 1375 mmol/L",
    "SUCB": "0.292 - 40.9 mmol/L",
    "SUCD": "2.92 - 409 mmol/L",
    "SUGLB": "0.292 - 40.9 mmol/L",
    "SUGLD": "2.92 - 409 mmol/L",
    "TP2B": "4 - 120 g/L",
    "TP2D": "40 - 1200 g/L",
    "TP2LB": "0.3 - 6.0 g/L",
}

SAMPLE_ROLE_TYPES = {"SAM": "Sample"}

SOLUTION_ANALYZER = "solution-analyzer"
OPTICAL_DENSITY = "optical_density"
BELOW_TEST_RANGE = "< TEST RNG"

FLAG_TO_ERROR = defaultdict(
    lambda: "unknown",
    {
        "v": "verification",
        "c": "calibration",
        BELOW_TEST_RANGE: BELOW_TEST_RANGE,
    },
)



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from enum import Enum
import logging
from pathlib import Path
import re
from typing import Any

import pandas as pd

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    ScanPositionSettingPlateReader,
    TransmittedLightSetting,
)
from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    DataSource,
    ImageFeature,
    ImageSource,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedData,
)
from allotropy.exceptions import AllotropeConversionError, AllotropeParsingError
from allotropy.parsers.lines_reader import CsvReader, EMPTY_STR_OR_CSV_LINE
from allotropy.parsers.revvity_kaleido import constants
from allotropy.parsers.utils.pandas import df_to_series_data, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float,
    try_float_or_none,
)


class ExperimentType(Enum):
    FLUORESCENCE = "fluorescence"
    ABSORBANCE = "absorbance"
    LUMINESCENCE = "luminescence"
    OPTICAL_IMAGING = "optical imaging"

    @staticmethod
    def create(experiment_type: str) -> ExperimentType:
        experiment_type_lower = experiment_type.lower()
        if "fluorescence" in experiment_type_lower or "alpha" in experiment_type_lower:
            return ExperimentType.FLUORESCENCE
        elif "abs" in experiment_type_lower:
            return ExperimentType.ABSORBANCE
        elif "luminescence" in experiment_type_lower:
            return ExperimentType.LUMINESCENCE
        elif "img" in experiment_type_lower:
            return ExperimentType.OPTICAL_IMAGING

        msg = f"Unable to determine experiment type from: '{experiment_type}'"
        raise AllotropeConversionError(msg)

    @property
    def measurement_type(self) -> MeasurementType:
        if self is ExperimentType.FLUORESCENCE:
            return MeasurementType.FLUORESCENCE
        elif self is ExperimentType.ABSORBANCE:
            return MeasurementType.ULTRAVIOLET_ABSORBANCE
        elif self is ExperimentType.LUMINESCENCE:
            return MeasurementType.LUMINESCENCE
        elif self is ExperimentType.OPTICAL_IMAGING:
            return MeasurementType.OPTICAL_IMAGING

    @property
    def device_type(self) -> str:
        if self is ExperimentType.OPTICAL_IMAGING:
            return "imaging detector"
        return f"{self.value} detector"

    @property
    def detection_type(self) -> str | None:
        # TODO(nstender): investigate why we override detection type for these two type but not others,
        # when we have detection type value for all examples.
        if self is ExperimentType.OPTICAL_IMAGING:
            return "optical-imaging"
        elif self is ExperimentType.FLUORESCENCE:
            return str(self.value)
        return None


class Version(Enum):
    V2 = "2.0"
    V3 = "3.0"


@dataclass
class BackgroundInfo:
    experiment_type_value: str

    @property
    def experiment_type(self) -> ExperimentType:
        return ExperimentType.create(self.experiment_type_value)

    @staticmethod
    def create(reader: CsvReader) -> BackgroundInfo:
        line = assert_not_none(
            reader.drop_until_inclusive("^Results? for.(.+) 1"),
            msg="Unable to find background information.",
        )

        experiment_type = assert_not_none(
            re.match("^Results? for.(.+) 1", line),
            msg="Unable to find experiment type from background information section.",
        ).group(1)

        return BackgroundInfo(experiment_type_value=experiment_type)


@dataclass(frozen=True)
class Results:
    barcode: str
    data: dict[str, str]

    @property
    def plate_well_count(self) -> int:
        return len(self.data)

    @classmethod
    def create(cls, reader: CsvReader) -> Results:
        barcode = cls.read_barcode(reader)

        results = assert_not_none(
            reader.pop_csv_block_as_df(
                header=0, index_col=0, empty_pat=EMPTY_STR_OR_CSV_LINE
            ),
            msg="Unable to find results table.",
        )

        for column in results:
            if str(column).startswith("Unnamed"):
                results = results.drop(columns=column)

        return Results(
            barcode=barcode,
            data={
                f"{row}{col}": str(values[col])
                for row, values in results.iterrows()
                for col in results.columns
            },
        )

    @classmethod
    def read_barcode(cls, reader: CsvReader) -> str:
        raise NotImplementedError


@dataclass(frozen=True)
class AnalysisResult:
    analysis_parameter: str
    results: dict[str, str]

    @staticmethod
    def create_results(
        reader: CsvReader, next_section_title: str
    ) -> list[AnalysisResult]:
        section_title = assert_not_none(
            reader.drop_until(f"^Results? for|^{next_section_title}"),
            msg=f"Unable to find Analysis Result or {next_section_title} section.",
        )

        if section_title.startswith(next_section_title):
            return []

        reader.drop_until("^Barcode")

        analysis_results = []
        while reader.match("^Barcode"):
            if analysis_result := AnalysisResult.create(reader):
                analysis_results.append(analysis_result)

        return analysis_results

    @staticmethod
    def create(reader: CsvReader) -> AnalysisResult | None:
        barcode_line = assert_not_none(
            reader.drop_until_inclusive("^Barcode:(.+),"),
            msg="Unable to find background information.",
        )

        analysis_parameter = None
        for element in barcode_line.split(","):
            if "AnalysisParameter" in element:
                analysis_parameter = element.split(":", maxsplit=1)[1]

        analysis_parameter = assert_not_none(
            analysis_parameter,
            msg="Unable to find analysis parameter in Analysis Results section.",
        )

        try:
            results_df = assert_not_none(
                reader.pop_csv_block_as_df(
                    header=0, index_col=0, empty_pat=EMPTY_STR_OR_CSV_LINE
                ),
                msg="Unable to find results table.",
            ).dropna(how="all")
        except AllotropeParsingError:
            logging.warning(
                f"Unable to read analysis result '{analysis_parameter}'. Ignoring"
            )
            return None

        results = {
            f"{row}{col}": str(values[col])
            for row, values in results_df.iterrows()
            for col in results_df.columns
        }

        # if first value is not a valid float value the results are not useful
        if try_float_or_none(results.get("A1")) is None:
            return None

        return AnalysisResult(
            analysis_parameter=analysis_parameter,
            results=results,
        )

    def get_image_feature_result(self, well_position: str) -> float:
        value_name = f"analysis result '{self.analysis_parameter}' at '{well_position}'"
        return try_float(
            assert_not_none(self.results.get(well_position), value_name), value_name
        )


@dataclass(frozen=True)
class MeasurementInfo:
    instrument_serial_number: str
    measurement_time: str
    protocol_signature: str
    measurement_signature: str
    custom_info: dict[str, Any]

    @staticmethod
    def create(
        reader: CsvReader, section_name: str, next_section_title: str
    ) -> MeasurementInfo:
        assert_not_none(
            reader.drop_until_inclusive(f"^{section_name}"),
            msg=f"Unable to find {section_name} section.",
        )

        lines = list(reader.pop_until(f"^{next_section_title}"))
        # Because we may read over multiple sections of data to get all metadata for this section, there may
        # be varying number of columns. Get the max number of columns, so we can read in all sections without error.
        max_num_cols = max(len(line.split(",")) for line in lines)
        df = assert_not_none(
            reader.lines_as_df(lines, index_col=0, names=range(max_num_cols)),
            msg=f"Unable to parser data for {section_name} section.",
        ).T.dropna(how="all")
        df.columns = df.columns.astype(str).str.lower().str.strip(":")
        data = df_to_series_data(df, index=0)

        return MeasurementInfo(
            instrument_serial_number=data[str, "instrument serial number"],
            measurement_time=data[str, "measurement started"],
            protocol_signature=data[str, "protocol signature"],
            measurement_signature=data[str, "measurement signature"],
            custom_info=data.get_unread(skip={"software version", "barcode", "nan"}),
        )


@dataclass(frozen=True)
class Platemap:
    data: dict[str, str]

    @staticmethod
    def create(reader: CsvReader) -> Platemap:
        assert_not_none(
            reader.drop_until_inclusive("^Platemap"),
            msg="Unable to find Platemap section.",
        )
        reader.pop_if_match("^Plate")

        data = assert_not_none(
            reader.pop_csv_block_as_df(
                header=0, index_col=0, empty_pat=EMPTY_STR_OR_CSV_LINE
            ),
            msg="Unable to find platemap information.",
        )

        return Platemap(
            data={
                f"{row}{col}": str(values[col])
                for row, values in data.iterrows()
                for col in data.columns
            }
        )

    def get(self, well_position: str) -> str:
        return assert_not_none(
            self.data.get(well_position),
            f"platemap value for well position '{well_position}'",
        )

    def get_sample_role_type(self, well_position: str) -> SampleRoleType | None:
        raw_value = self.get(well_position)
        if raw_value == "-":
            return None

        value = assert_not_none(
            re.match(r"^([A-Z]+)\d+$", raw_value),
            msg=f"Unable to understand platemap value '{raw_value}' for well position '{well_position}'.",
        ).group(1)

        return assert_not_none(
            constants.PLATEMAP_TO_SAMPLE_ROLE_TYPE.get(value),
            msg=f"Unable to find sample role type for well position '{well_position}'.",
        )


@dataclass(frozen=True)
class Channel:
    name: str
    fluorescent_tag: str | None
    transmitted_light: TransmittedLightSetting | None
    excitation_wavelength: float
    illumination: float
    exposure_duration: float
    additional_focus_offset: float

    @staticmethod
    def create(
        name: str,
        excitation_wavelength: str,
        excitation_power: str,
        exposure_time: str,
        additional_focus_offset: str,
    ) -> Channel:
        return Channel(
            name=name,
            fluorescent_tag=None if name == "BRIGHTFIELD" else name,
            transmitted_light=constants.TRANSMITTED_LIGHT_CONVERSION.get(name),
            excitation_wavelength=try_float(
                excitation_wavelength.removesuffix("nm"), "excitation wavelength"
            ),
            illumination=try_float(excitation_power, "excitation power"),
            exposure_duration=try_float(exposure_time, "exposure time"),
            additional_focus_offset=try_float(
                additional_focus_offset, "additional focus offset"
            ),
        )


@dataclass(frozen=True)
class Measurements:
    channels: list[Channel]
    number_of_averages: float | None
    detector_distance_luminiscence: float | None
    detector_distance: float | None
    scan_position: ScanPositionSettingPlateReader | None
    emission_wavelength: float | None
    excitation_wavelength: float | None
    focus_height: float | None
    device_control_custom_info: dict[str, Any]
    custom_info: dict[str, Any] | None = None

    @staticmethod
    def create(
        reader: CsvReader,
        section_title: str,
        next_section_title: str,
        measurement_info: MeasurementInfo,
    ) -> Measurements:
        assert_not_none(
            reader.drop_until_inclusive(f"^{section_title}"),
            msg=f"Unable to find {section_title} section.",
        )

        lines = list(reader.pop_until(f"^{next_section_title}"))
        # Because we may read over multiple sections of data to get all metadata for this section, there may
        # be varying number of columns. Get the max number of columns, so we can read in all sections without error.
        max_num_cols = max(len(line.split(",")) for line in lines)
        df = assert_not_none(
            reader.lines_as_df(lines, index_col=0, names=range(max_num_cols)),
            msg=f"Unable to parser data for {section_title} section.",
        ).T.dropna(how="all")
        df.columns = df.columns.astype(str).str.lower()
        data = df_to_series_data(df, index=0)

        scan_position = data.get(str, "excitation / emission")
        excitation_wavelength = data.get(str, "excitation wavelength [nm]")

        return Measurements(
            channels=Measurements.create_channels(data),
            number_of_averages=data.get(float, "number of flashes"),
            detector_distance=data.get(float, "measurement height [mm]"),
            detector_distance_luminiscence=data.get(
                float, "distance between plate and detector [mm]"
            ),
            scan_position=(
                None
                if scan_position is None
                else assert_not_none(
                    constants.SCAN_POSITION_CONVERSION.get(scan_position),
                    msg=f"'{scan_position}' is not a valid scan position, expected TOP or BOTTOM.",
                )
            ),
            emission_wavelength=data.get(float, "emission wavelength [nm]"),
            excitation_wavelength=(
                None
                if excitation_wavelength is None
                else try_float_or_none(excitation_wavelength.removesuffix("nm"))
            ),
            focus_height=data.get(float, "focus height [µm]"),
            device_control_custom_info={
                "number of flashes integrated": data.get(
                    float, "number of flashes integrated"
                ),
                "flash power": data.get(float, "flash power"),
                "measurement mode setting": data.get(str, "measurement mode"),
                "start plate repeat each [s]": measurement_info.custom_info.get(
                    "start plate repeat each [s]", None
                ),
                "number of plate repeats": measurement_info.custom_info.get(
                    "number of plate repeats", None
                ),
                "sequence executed by": measurement_info.custom_info.get(
                    "sequence executed by", None
                ),
                "filter set": data.get(str, "filter set"),
                "generate bright field": data.get(bool, "generate bright field"),
                "digital phase contrast": data.get(bool, "digital phase contrast"),
                "generate dpc": data.get(bool, "generate dpc"),
            },
            custom_info=data.get_unread(
                skip={
                    "tech",
                    "operation",
                    "nan",
                }
            ),
        )

    @staticmethod
    def create_channels(data: SeriesData) -> list[Channel]:
        # Get all channel keys
        channel_keys = [
            "channel",
            "excitation wavelength [nm]",
            "excitation power [%]",
            "exposure time [ms]",
            "additional focus offset [mm]",
        ]
        values = {key: data.series.get(key) for key in channel_keys}
        data.mark_read(set(channel_keys))
        if values["channel"] is None:
            return []

        # Convert series values (multiple Channels == pd.Series, single Channel == str) to lists.
        channel_values: dict[str, list[str]] = {}
        if isinstance(values["channel"], str):
            for key, value in values.items():
                if not isinstance(value, str):
                    raise AllotropeConversionError(constants.CHANNEL_COLUMNS_ERROR)
                channel_values[key] = [value]
        elif isinstance(values["channel"], pd.Series):
            for key, value in values.items():
                if not isinstance(value, pd.Series):
                    raise AllotropeConversionError(constants.CHANNEL_COLUMNS_ERROR)
                channel_values[key] = list(value.values)
                if not len(channel_values["channel"]) == len(channel_values[key]):
                    raise AllotropeConversionError(constants.CHANNEL_COLUMNS_ERROR)

        return [
            Channel.create(
                name=channel_values["channel"][i],
                excitation_wavelength=channel_values["excitation wavelength [nm]"][i],
                excitation_power=channel_values["excitation power [%]"][i],
                exposure_time=channel_values["exposure time [ms]"][i],
                additional_focus_offset=channel_values["additional focus offset [mm]"][
                    i
                ],
            )
            for i in range(len(channel_values["channel"]))
        ]


@dataclass(frozen=True)
class Data:
    version: str
    background_info: BackgroundInfo
    results: Results
    analysis_results: list[AnalysisResult]
    measurement_info: MeasurementInfo
    platemap: Platemap
    measurements: Measurements

    def iter_wells(self) -> Iterator[tuple[str, str]]:
        yield from self.results.data.items()

    def get_sample_identifier(self, well_position: str) -> str:
        platemap_value = self.platemap.get(well_position)
        return (
            f"{self.results.barcode}_{well_position}"
            if platemap_value in ("-", None)
            else platemap_value
        )


def create_metadata(data: Data, file_path: str) -> Metadata:
    return Metadata(
        file_name=Path(file_path).name,
        unc_path=file_path,
        software_name=constants.SOFTWARE_NAME,
        software_version=data.version,
        device_identifier=constants.DEVICE_IDENTIFIER,
        model_number=constants.MODEL_NUMBER,
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        equipment_serial_number=data.measurement_info.instrument_serial_number,
        custom_info=data.measurement_info.custom_info,
    )


def _create_measurement(data: Data, well_position: str, well_value: str) -> Measurement:
    experiment_type = data.background_info.experiment_type

    if experiment_type is ExperimentType.ABSORBANCE:
        assert_not_none(
            data.measurements.number_of_averages,
            msg="Unable to find number of averages",
        )

    detector_wavelength_setting = (
        data.measurements.emission_wavelength
        if experiment_type is ExperimentType.FLUORESCENCE
        else (
            data.measurements.excitation_wavelength
            if experiment_type is ExperimentType.ABSORBANCE
            else None
        )
    )
    measurement_value = try_float(well_value, f"result well at '{well_position}'")

    return Measurement(
        type_=experiment_type.measurement_type,
        identifier=random_uuid_str(),
        device_type=experiment_type.device_type,
        detection_type=experiment_type.detection_type
        or data.background_info.experiment_type_value,
        sample_identifier=data.get_sample_identifier(well_position),
        location_identifier=well_position,
        well_plate_identifier=data.results.barcode,
        sample_role_type=data.platemap.get_sample_role_type(well_position),
        number_of_averages=data.measurements.number_of_averages,
        detector_distance_setting=data.measurements.detector_distance_luminiscence
        if experiment_type is ExperimentType.LUMINESCENCE
        else data.measurements.detector_distance,
        scan_position_setting=data.measurements.scan_position,
        detector_wavelength_setting=detector_wavelength_setting,
        excitation_wavelength_setting=data.measurements.excitation_wavelength,
        fluorescence=measurement_value
        if experiment_type is ExperimentType.FLUORESCENCE
        else None,
        absorbance=measurement_value
        if experiment_type is ExperimentType.ABSORBANCE
        else None,
        luminescence=measurement_value
        if experiment_type is ExperimentType.LUMINESCENCE
        else None,
        device_control_custom_info=data.measurements.device_control_custom_info,
        custom_info=data.measurements.custom_info,
    )


def _create_optical_measurement(
    data: Data, well_position: str, channel: Channel
) -> Measurement:
    device_control_custom_info = {
        **data.measurements.device_control_custom_info,
        "additional focus offset [mm]": channel.additional_focus_offset,
    }
    if channel.name != "BRIGHTFIELD":
        to_remove = ["generate bright field", "digital phase contrast", "generate dpc"]
        for key in to_remove:
            device_control_custom_info.pop(key, None)

    return Measurement(
        type_=data.background_info.experiment_type.measurement_type,
        identifier=random_uuid_str(),
        device_type=data.background_info.experiment_type.device_type,
        detection_type=data.background_info.experiment_type.detection_type
        or data.background_info.experiment_type_value,
        sample_identifier=data.get_sample_identifier(well_position),
        location_identifier=well_position,
        well_plate_identifier=data.results.barcode,
        sample_role_type=data.platemap.get_sample_role_type(well_position),
        detector_distance_setting=data.measurements.focus_height,
        excitation_wavelength_setting=channel.excitation_wavelength,
        magnification_setting=constants.MAGNIFICATION_SETTING,
        exposure_duration_setting=channel.exposure_duration,
        illumination_setting=channel.illumination,
        illumination_setting_unit="%",
        transmitted_light_setting=channel.transmitted_light,
        fluorescent_tag_setting=channel.fluorescent_tag,
        device_control_custom_info=device_control_custom_info,
        custom_info=data.measurements.custom_info,
    )


def _create_measurements(
    data: Data, well_position: str, well_value: str
) -> list[Measurement]:
    if data.background_info.experiment_type is ExperimentType.OPTICAL_IMAGING:
        return [
            _create_optical_measurement(data, well_position, channel)
            for channel in data.measurements.channels
        ]
    else:
        return [_create_measurement(data, well_position, well_value)]


def create_measurement_groups(data: Data) -> list[MeasurementGroup]:
    protocol_owner = data.measurement_info.custom_info.pop("protocol owner", None)
    measurement_finished = data.measurement_info.custom_info.pop(
        "measurement finished", None
    )
    protocol_name = data.measurement_info.custom_info.pop("protocol name", None)

    data.measurement_info.custom_info.pop("sequence executed by", None)
    data.measurement_info.custom_info.pop("number of plate repeats", None)
    data.measurement_info.custom_info.pop("start plate repeat each [s]", None)

    return [
        MeasurementGroup(
            measurement_time=data.measurement_info.measurement_time,
            plate_well_count=data.results.plate_well_count,
            experiment_type=data.background_info.experiment_type_value,
            analytical_method_identifier=data.measurement_info.protocol_signature,
            experimental_data_identifier=data.measurement_info.measurement_signature,
            measurements=_create_measurements(data, well_position, well_value),
            processed_data=ProcessedData(
                features=[
                    ImageFeature(
                        identifier=random_uuid_str(),
                        feature=analysis_result.analysis_parameter,
                        result=analysis_result.get_image_feature_result(well_position),
                        data_sources=[
                            DataSource(
                                identifier=well_value,
                                # TODO(nstender): I'm not sure what this original comment means, investigate.
                                feature="image feature",  # wait for actual value
                            )
                        ],
                    )
                    for analysis_result in data.analysis_results
                ]
            ),
            images=[
                ImageSource(identifier=well_value),
            ]
            if data.background_info.experiment_type is ExperimentType.OPTICAL_IMAGING
            else None,
            custom_info={
                "protocol owner": protocol_owner,
                "measurement finished": measurement_finished,
                "protocol name": protocol_name,
            },
        )
        for well_position, well_value in data.iter_wells()
    ]



from allotropy.calcdocs.extractor import Element
from allotropy.calcdocs.perkin_elmer_envision.extractor import (
    PerkinElmerEnvisionExtractor,
)
from allotropy.calcdocs.view import View, ViewData
from allotropy.parsers.perkin_elmer_envision.perkin_elmer_envision_structure import (
    CalculatedPlate,
    PlateList,
)


class PosView(View):
    def __init__(self, sub_view: View | None = None):
        super().__init__(name="pos", sub_view=sub_view)

    def sort_elements(self, _: list[Element]) -> dict[str, list[Element]]:
        return {}

    def special_apply(
        self, plate_list: PlateList, calculated_plate: CalculatedPlate
    ) -> ViewData:
        return ViewData(
            view=self,
            name=self.name,
            data={
                calculated_result.pos: [
                    PerkinElmerEnvisionExtractor.to_element_with_extra_data(
                        source_result,
                        extra_data={"calc value": calculated_result.value},
                    )
                    for source_result in source_results
                ]
                for calculated_result, source_results in calculated_plate.get_result_and_sources(
                    plate_list
                )
            },
        )



from io import StringIO

import pandas as pd

from allotropy.exceptions import AllotropeConversionError, AllotropeParsingError
from allotropy.parsers.utils.pandas import assert_not_empty_df, read_csv
from allotropy.types import IOType


class NucleoviewReader:
    SUPPORTED_EXTENSIONS = "csv"

    @classmethod
    def read(cls, contents: IOType) -> pd.DataFrame:
        content = contents.read()
        content_str = content.decode("UTF-8") if isinstance(content, bytes) else content
        separators = [":\t", "[;,]+"]
        last_error = None

        for sep in separators:
            try:
                df = read_csv(
                    StringIO(content_str),
                    sep=sep,
                    engine="python",
                    skipinitialspace=True,
                    index_col=0,
                )
                if df.shape[1] >= 1:
                    break
            except AllotropeParsingError as e:
                last_error = e
                continue
        else:
            if last_error:
                raise last_error

        try:
            df = df[:-1].dropna(axis="index", how="all").T
        except KeyError as e:
            msg = f"Error processing CSV data structure. The file may be corrupted or have an invalid format: {e}"
            raise AllotropeParsingError(msg) from e

        df = df.rename(
            {"Estimated cell diameter [um]": "Estimated cell diameter (um)"},
            axis="columns",
        )
        if "Viability (%)" not in df.columns:
            msg = "Value for Cell Viability was expected, but was not found in the input file"
            raise AllotropeConversionError(msg)
        df = df.dropna(subset=["Viability (%)"])
        df = df[df.columns.dropna()]
        return assert_not_empty_df(df, "Unable to parse data from empty dataset.")



from __future__ import annotations

from pathlib import Path

import pandas as pd

from allotropy.allotrope.models.shared.definitions.units import (
    Micrometer,
    Percent,
)
from allotropy.allotrope.schema_mappers.adm.cell_counting.rec._2024._09.cell_counting import (
    CalculatedDataItem,
    Error,
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.parsers.chemometec_nucleoview.constants import (
    DEFAULT_ANALYST,
    DEFAULT_MODEL_NUMBER,
    MEASUREMENT_AGG_DOCUMENT_CUSTOM_FIELDS,
    NUCLEOCOUNTER_DETECTION_TYPE,
    NUCLEOCOUNTER_DEVICE_TYPE,
    NUCLEOCOUNTER_SOFTWARE_NAME,
)
from allotropy.parsers.constants import (
    DEFAULT_EPOCH_TIMESTAMP,
    NEGATIVE_ZERO,
    NOT_APPLICABLE,
)
from allotropy.parsers.utils.calculated_data_documents.definition import (
    DataSource,
    Referenceable,
)
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def create_metadata(data: SeriesData, file_path: str) -> Metadata:
    path = Path(file_path)
    metadata = Metadata(
        asm_file_identifier=path.with_suffix(".json").name,
        data_system_instance_id=data.get(str, "PC", NOT_APPLICABLE),
        file_name=Path(file_path).name,
        unc_path=file_path,
        model_number=data.get(str, "Instrument type", DEFAULT_MODEL_NUMBER),
        equipment_serial_number=data.get(str, "Instrument s/n"),
        software_name=NUCLEOCOUNTER_SOFTWARE_NAME,
        software_version=data.get(str, "Application SW version"),
        device_type=NUCLEOCOUNTER_DEVICE_TYPE,
        detection_type=NUCLEOCOUNTER_DETECTION_TYPE,
        data_system_custom_info=data.get_custom_keys(
            {
                "csv file version",
                "21 CFR Part 11",
            }
        ),
    )
    # We read header info from a row in the table, so we don't need to read all keys from this SeriesData
    data.get_unread()
    return metadata


def create_measurement_groups(
    data: SeriesData,
) -> tuple[MeasurementGroup, list[CalculatedDataItem] | None]:
    timestamp = data.get(str, "Date time")
    errors = []
    if timestamp:
        offset = data.get(str, "Time zone offset", "0")
        timestamp = pd.to_datetime(
            f"{timestamp}{'+' if offset[0] not in {'+', '-'} else ''}{offset}"
        ).isoformat()

    def _converted_value_or_none(key: str) -> float | None:
        return value / 1e6 if (value := data.get(float, key)) is not None else None

    viable_cell_density = data.get(float, "Live (cells/ml)")
    if viable_cell_density is None:
        errors.append(
            Error(
                error=NOT_APPLICABLE,
                feature="viable cell density",
            )
        )
    # These fields we read from the first row as metadata, so we don't need to read them here
    data.mark_read(
        {
            "PC",
            "Instrument type",
            "Instrument s/n",
            "Application SW version",
            "21 CFR Part 11",
            "csv file version",
        }
    )
    percentage_of_cells_with_five_or_more = data.get(
        float, "(%) of cells in aggregates with five or more cells"
    )
    cell_diameter_standard_deviation = data.get(
        float, "Cell diameter standard deviation (um)"
    )
    measurement_group = MeasurementGroup(
        analyst=data.get(str, "Operator", DEFAULT_ANALYST),
        custom_info=data.get_custom_keys(MEASUREMENT_AGG_DOCUMENT_CUSTOM_FIELDS),
        measurements=[
            Measurement(
                measurement_identifier=random_uuid_str(),
                timestamp=timestamp or DEFAULT_EPOCH_TIMESTAMP,
                sample_identifier=data.get(str, "Sample ID")
                or data[str, "Image"].split("-")[3],
                cell_density_dilution_factor=data.get(float, "Multiplication factor"),
                viability=data[float, "Viability (%)"],
                # Cell counts are measured in cells/mL, but reported in millions of cells/mL
                viable_cell_density=(viable_cell_density / 1e6)
                if viable_cell_density
                else NEGATIVE_ZERO,
                dead_cell_density=_converted_value_or_none("Dead (cells/ml)"),
                total_cell_density=_converted_value_or_none("Total (cells/ml)"),
                average_total_cell_diameter=data.get(
                    float, "Estimated cell diameter (um)"
                ),
                errors=errors,
                sample_volume_setting=data.get(float, "Sample Volume (ul)"),
                experimental_data_identifier=data.get(str, "cm filename"),
                dilution_volume=data.get(float, "Dilution Volume (ul)"),
                custom_info=data.get_unread(),
            )
        ],
    )

    return measurement_group, _get_calculated_data(
        [measurement_group],
        cell_diameter_standard_deviation,
        percentage_of_cells_with_five_or_more,
    )


def _get_calculated_data(
    groups: list[MeasurementGroup],
    cell_diameter_standard_deviation: float | None,
    percentage_of_cells_with_five_or_more: float | None,
) -> list[CalculatedDataItem] | None:
    result = []
    for group in groups:
        if cell_diameter_standard_deviation:
            result.append(
                CalculatedDataItem(
                    identifier=random_uuid_str(),
                    name="Cell diameter standard deviation (um)",
                    value=cell_diameter_standard_deviation,
                    data_sources=[
                        DataSource(
                            feature="Average Total Cell Diameter",
                            reference=Referenceable(
                                uuid=group.measurements[0].measurement_identifier
                            ),
                        )
                    ],
                    unit=Micrometer.unit,
                )
            )
        if percentage_of_cells_with_five_or_more:
            result.append(
                CalculatedDataItem(
                    identifier=random_uuid_str(),
                    name="Percentage of cells with five or more",
                    value=percentage_of_cells_with_five_or_more,
                    data_sources=[
                        DataSource(
                            feature="Cell count",
                            reference=Referenceable(
                                uuid=group.measurements[0].measurement_identifier
                            ),
                        )
                    ],
                    unit=Percent.unit,
                )
            )
    return result
